"""
About GigaST corpus:
GigaST: A large-scale speech translation corpus
https://arxiv.org/abs/2204.03939

GigaST is a large-scale speech translation corpus, by translating the transcriptions in GigaSpeech, a multi-domain English speech recognition corpus with 10,000 hours of labeled audio. The training data is translated by a strong machine translation system and the test data is produced by professional human translators.

Note:
This recipe assume you have downloaded and prepared GigaSpeech by lhotse.
We assure manifests_dir contains the following files:
  - gigaspeech_recordings_TEST.jsonl.gz
  - gigaspeech_recordings_XL.jsonl.gz
  - gigaspeech_supervisions_TEST.jsonl.gz
  - gigaspeech_supervisions_XL.jsonl.gz
"""
import json
import logging
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse import CutSet
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.recipes.utils import manifests_exist, read_manifests_if_cached
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, Seconds, is_module_available, resumable_download

GIGASPEECH_PARTS = ("XL", "L", "M", "S", "XS", "DEV", "TEST")
GIGAST_LANGS = ("de", "zh")


class GigaST:
    def __init__(self, corpus_dir: Pathlike, lang: str):
        with open(corpus_dir / f"GigaST.{lang}.json") as f:
            self.audio_generator = iter(json.load(f)["audios"])
        self.segment_generator = iter(next(self.audio_generator)["segments"])

    def get_next_line(self):
        try:
            return next(self.segment_generator)
        except StopIteration:
            self.segment_generator = iter(next(self.audio_generator)["segments"])
            return next(self.segment_generator)


def download_gigast(
    target_dir: Pathlike = ".",
    languages: Union[str, Sequence[str]] = "all",
    force_download: bool = False,
) -> Path:
    """
    Download GigaST dataset

    :param target_dir: Pathlike, the path of the dir to store the dataset.
    :param languages: one of: 'all' (downloads all known languages); a single language code (e.g., 'en'), or a list of language codes.
    :param force_download: bool, if True, download the archive even if it already exists.

    :return: the path to downloaded with data.
    """
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    if languages == "all":
        languages = GIGAST_LANGS
    elif isinstance(languages, str):
        languages = [languages]
    else:
        languages = list(languages)

    for lang in tqdm(languages, desc=f"Downloading GigaST"):
        logging.info(f"Download language: {lang}")
        completed_detector = target_dir / f".{lang}_completed"
        if completed_detector.is_file():
            logging.info(f"Skipping {lang} because {completed_detector} exists.")
            continue
        # Process the archive.
        json_name = f"GigaST.{lang}.json"
        json_path = target_dir / json_name
        resumable_download(
            f"https://lf3-nlp-opensource.bytetos.com/obj/nlp-opensource/datasets/GigaST/{json_name}",
            filename=json_path,
            force_download=force_download,
        )
        completed_detector.touch()

    return target_dir


def prepare_gigast(
    corpus_dir: Pathlike,
    manifests_dir: Pathlike,
    output_dir: Optional[Pathlike],
    languages: Union[str, Sequence[str]] = "auto",
    dataset_parts: Union[str, Sequence[str]] = "auto",
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Path to the GigaST dataset
    :param manifests_dir: Path to the GigaSpeech manifests
    :param output_dir: Pathlike, the path where to write the manifests.
    :param languages: 'auto' (prepare all languages) or a list of language codes.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    manifests_dir = Path(manifests_dir)

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    logging.info("Preparing GigaST...")

    languages = GIGAST_LANGS if languages == "auto" else languages
    if isinstance(languages, str):
        languages = [languages]

    dataset_parts = ("XL", "TEST") if dataset_parts == "auto" else dataset_parts
    if isinstance(dataset_parts, str):
        dataset_parts = [dataset_parts]

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    logging.info("Loading manifest")
    prefix = "gigaspeech"
    suffix = "jsonl.gz"
    manifests = read_manifests_if_cached(
        dataset_parts=dataset_parts,
        output_dir=manifests_dir,
        prefix=prefix,
        suffix=suffix,
    )
    assert manifests is not None

    assert len(manifests) == len(dataset_parts), (
        len(manifests),
        len(dataset_parts),
        list(manifests.keys()),
        dataset_parts,
    )

    for lang in tqdm(languages, desc="Processing GigaST"):
        assert lang in GIGAST_LANGS, (lang, GIGAST_LANGS)
        logging.info(f"Loading GigaST.{lang}.json")
        gigast = GigaST(corpus_dir, lang)

        for partition, m in manifests.items():
            supervisions = []
            logging.info(f"Processing {partition}")
            if manifests_exist(
                part=partition,
                output_dir=output_dir,
                prefix="gigast-de",
                suffix="jsonl.gz",
            ):
                logging.info(
                    f"GigaST {lang} subset: {partition} already prepared - skipping."
                )
                continue

            cur_line = gigast.get_next_line()
            for sup in tqdm(
                m["supervisions"], desc="Generate the extented supervisions"
            ):
                if cur_line["sid"] == sup.id:
                    new_sup = sup
                    if partition != "TEST":
                        new_sup.custom = {
                            "text_raw": cur_line["text_raw"],
                            "extra": cur_line["extra"],
                        }
                    else:
                        new_sup.custom = {"text_raw": cur_line["text_raw"]}
                    supervisions.append(new_sup)
                    try:
                        cur_line = gigast.get_next_line()
                    except StopIteration:
                        break

            logging.info(f"Saving GigaST {lang} subset: {partition}")
            supervisionset = SupervisionSet.from_segments(supervisions)
            supervisionset.to_file(
                output_dir / f"gigast-{lang}_supervisions_{partition}.jsonl.gz"
            )



"""
Multi-Domain Cantonese Corpus (MDCC), consists of 73.6 hours of clean read speech paired with 
transcripts, collected from Cantonese audiobooks from Hong Kong. It comprises philosophy, 
politics, education, culture, lifestyle and family domains, covering a wide range of topics. 

Manuscript can be found at: https://arxiv.org/abs/2201.02419
"""

import logging
import zipfile
from pathlib import Path
from typing import Dict, Sequence, Union

from tqdm.auto import tqdm

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available

MDCC_URL = "https://drive.google.com/file/d/1epfYMMhXdBKA6nxPgUugb2Uj4DllSxkn/view"

MDCC_PARTS = ["train", "valid", "test"]


def download_mdcc(target_dir: Pathlike, force_download: bool = False) -> Path:
    """
    Downloads the MDCC data from the Google Drive and extracts it.
    :param target_dir: the directory where MDCC data will be saved.
    :param force_download: if True, it will download the MDCC data even if it is already present.
    :return: the path to downloaded and extracted directory with data.
    """
    if not is_module_available("gdown"):
        raise ValueError("Please run 'pip install gdown' to download MDCC.")

    import gdown

    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    corpus_dir = target_dir / "dataset"
    corpus_zip = corpus_dir.with_suffix(".zip")

    if not force_download and corpus_zip.exists():
        logging.info(f"{corpus_zip} already exists. Skipping download.")
    else:
        logging.info(f"Running: gdown --fuzzy {MDCC_URL}")
        gdown.download(MDCC_URL, str(corpus_zip), fuzzy=True, quiet=False)

    # Extract the zipped file
    if not corpus_dir.exists() or force_download:
        logging.info(f"Extracting {corpus_zip} to {target_dir}")
        with zipfile.ZipFile(corpus_zip) as zf:
            zf.extractall(path=target_dir)

    return corpus_dir


def prepare_mdcc(
    corpus_dir: Pathlike,
    dataset_parts: Union[str, Sequence[str]] = "all",
    output_dir: Pathlike = None,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Create RecordingSet and SupervisionSet manifests for MDCC from a raw corpus distribution.

    :param corpus_dir: Pathlike, the path to the extracted corpus.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    audio_dir = corpus_dir / "audio"
    assert (audio_dir).is_dir(), f"Missing {audio_dir} in {corpus_dir}."
    manifests = {}

    if dataset_parts == "all" or dataset_parts[0] == "all":
        dataset_parts = MDCC_PARTS
    elif isinstance(dataset_parts, str):
        assert dataset_parts in MDCC_PARTS, f"Unknown dataset part: {dataset_parts}"
        dataset_parts = [dataset_parts]

    for part in dataset_parts:
        recordings = []
        supervisions = []

        metadata = corpus_dir / f"cnt_asr_{part}_metadata.csv"
        assert (metadata).is_file(), f"Missing {part} metadata in {corpus_dir}."

        # read cvs file in an ugly way as there are no more than 80k lines
        # and i don't want to depend on pandas
        with open(metadata, "r") as f:
            lines = f.readlines()

            # remove the header
            lines = lines[1:]

        for line in tqdm(lines, desc=f"Processing {part} metadata"):
            # audio_path, text_path, gender, duration
            audio_path, text_path, gender, _ = line.strip().split(",")
            audio_path = audio_dir / Path(audio_path).name
            text_path = corpus_dir / text_path

            recording_id = make_recording_id(Path(audio_path))
            recording = Recording.from_file(audio_path, recording_id=recording_id)
            recordings.append(recording)

            supervision_segment = SupervisionSegment(
                id=recording_id,
                recording_id=recording_id,
                start=0.0,
                duration=recording.duration,
                channel=0,
                text=text_path.read_text().strip(),
                gender=gender,
                language="yue",
            )
            supervisions.append(supervision_segment)

        recordings = RecordingSet.from_recordings(recordings)
        supervisions = SupervisionSet.from_segments(supervisions)
        recordings, supervisions = fix_manifests(recordings, supervisions)
        validate_recordings_and_supervisions(
            recordings=recordings, supervisions=supervisions
        )

        if output_dir is not None:
            output_dir = Path(output_dir)
            output_dir.mkdir(exist_ok=True, parents=True)
            recordings.to_file(output_dir / f"mdcc_recordings_{part}.jsonl.gz")
            supervisions.to_file(output_dir / f"mdcc_supervisions_{part}.jsonl.gz")

        manifests[part] = {"recordings": recordings, "supervisions": supervisions}

    return manifests


def make_recording_id(path: Path) -> str:
    return f"mdcc_{path.stem}"



import click

from lhotse.bin.modes import prepare
from lhotse.recipes.tal_asr import prepare_tal_asr
from lhotse.utils import Pathlike

__all__ = ["tal_asr"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def tal_asr(corpus_dir: Pathlike, output_dir: Pathlike):
    """Tal_asr ASR data preparation."""
    prepare_tal_asr(corpus_dir, output_dir=output_dir)



import glob
import json
import logging
from collections import defaultdict
from itertools import groupby
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

from tqdm import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, Seconds, is_module_available


def prepare_slu(
    corpus_dir: Pathlike, output_dir: Optional[Pathlike] = None
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:

    import pandas

    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    data = {
        "train": pandas.read_csv(
            str(corpus_dir) + "/data/train_data.csv", index_col=0, header=0
        ),
        "valid": pandas.read_csv(
            str(corpus_dir) + "/data/valid_data.csv", index_col=0, header=0
        ),
        "test": pandas.read_csv(
            str(corpus_dir) + "/data/test_data.csv", index_col=0, header=0
        ),
    }
    train_wavs = [
        str(corpus_dir) + "/" + path_to_wav
        for path_to_wav in data["train"]["path"].tolist()
    ]
    valid_wavs = [
        str(corpus_dir) + "/" + path_to_wav
        for path_to_wav in data["valid"]["path"].tolist()
    ]
    test_wavs = [
        str(corpus_dir) + "/" + path_to_wav
        for path_to_wav in data["test"]["path"].tolist()
    ]

    transcripts = {
        "train": data["train"]["transcription"].tolist(),
        "valid": data["valid"]["transcription"].tolist(),
        "test": data["test"]["transcription"].tolist(),
    }

    frames = {
        "train": list(
            i
            for i in zip(
                data["train"]["action"].tolist(),
                data["train"]["object"].tolist(),
                data["train"]["location"].tolist(),
            )
        ),
        "valid": list(
            i
            for i in zip(
                data["valid"]["action"].tolist(),
                data["valid"]["object"].tolist(),
                data["valid"]["location"].tolist(),
            )
        ),
        "test": list(
            i
            for i in zip(
                data["test"]["action"].tolist(),
                data["test"]["object"].tolist(),
                data["test"]["location"].tolist(),
            )
        ),
    }

    manifests = defaultdict(dict)
    for name, dataset in zip(
        ["train", "valid", "test"], [train_wavs, valid_wavs, test_wavs]
    ):
        recordings = []
        for wav in tqdm(dataset):
            recording = Recording.from_file(wav)
            recordings.append(recording)
        recording_set = RecordingSet.from_recordings(recordings)

        supervisions = []
        for id, recording in tqdm(enumerate(recording_set)):
            supervisions.append(
                SupervisionSegment(
                    id=id,
                    recording_id=recording.id,
                    start=0,
                    duration=recording.duration,
                    channel=0,
                    text=transcripts[name][id],
                    custom={"frames": frames[name][id]},
                )
            )
        supervision_set = SupervisionSet.from_segments(supervisions)
        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)
        manifests[name] = {
            "recordings": recording_set,
            "supervisions": supervision_set,
        }

    if output_dir is not None:
        for name in ["train", "valid", "test"]:
            manifests[name]["recordings"].to_file(
                output_dir / ("slu_recordings_" + name + ".jsonl.gz")
            )
            manifests[name]["supervisions"].to_file(
                output_dir / ("slu_supervisions_" + name + ".jsonl.gz")
            )



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.baker_zh import download_baker_zh, prepare_baker_zh
from lhotse.utils import Pathlike

__all__ = ["baker_zh"]


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path(), default=".")
def baker_zh(target_dir: Pathlike):
    """bazker_zh download."""
    download_baker_zh(target_dir)


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def baker_zh(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
):
    """bazker_zh data preparation."""
    prepare_baker_zh(corpus_dir, output_dir=output_dir)



"""
The ATCOSIM Air Traffic Control Simulation Speech corpus is a speech database of air traffic control (ATC) operator speech, provided by Graz University of Technology (TUG) and Eurocontrol Experimental Centre (EEC). It consists of ten hours of speech data, which were recorded during ATC real-time simulations using a close-talk headset microphone. The utterances are in English language and pronounced by ten non-native speakers. The database includes orthographic transcriptions and additional information on speakers and recording sessions. It was recorded an annotated by Konrad Hofbauer.

See https://www.spsc.tugraz.at/databases-and-tools/atcosim-air-traffic-control-simulation-speech-corpus.html for more details.
"""

import collections
import csv
import hashlib
import logging
import os
import re
import shutil
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import (
    Pathlike,
    Seconds,
    compute_num_samples,
    is_module_available,
    resumable_download,
)


# note: https://www2.spsc.tugraz.at/ does not support Range request header (2023-05-10)
def download_atcosim(
    target_dir: Pathlike = ".", force_download: Optional[bool] = False
) -> Path:
    if not is_module_available("pycdlib"):
        raise ImportError("Please 'pip install pycdlib' first.")
    import pycdlib

    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    dataset_name = "atcosim"
    iso_path = target_dir / f"{dataset_name}.iso"
    corpus_dir = target_dir / dataset_name
    completed_detector = corpus_dir / ".completed"
    if completed_detector.is_file():
        logging.info(f"Skipping {dataset_name} because {completed_detector} exists.")
        return corpus_dir
    resumable_download(
        f"https://www2.spsc.tugraz.at/databases/ATCOSIM/.ISO/{dataset_name}.iso",
        filename=iso_path,
        completed_file_size=2597789696,
        force_download=force_download,
    )
    if (
        hashlib.md5(open(iso_path, "rb").read()).hexdigest()
        != "cd5f0c82be46242a75d3382e340f6dca"
    ):
        raise RuntimeError("MD5 checksum does not match")
    shutil.rmtree(corpus_dir, ignore_errors=True)
    iso = pycdlib.PyCdlib()
    iso.open(iso_path)
    path_arg = (
        "udf"
        if iso.has_udf()
        else (
            "rr" if iso.has_rock_ridge() else ("joliet" if iso.has_joliet() else "iso")
        )
    ) + "_path"
    records = collections.deque([iso.get_record(**{path_arg: "/"})])
    while records:
        r = records.popleft()
        abs_path = iso.full_path_from_dirrecord(r, rockridge=path_arg == "rr_path")
        rel_path = abs_path.lstrip("/")
        if r.is_dir():
            os.makedirs(corpus_dir / rel_path)
            for child in iso.list_children(**{path_arg: abs_path}):
                if child and not (child.is_dot()) and not (child.is_dotdot()):
                    records.append(child)
        elif r.is_symlink():
            logging.warning("symlink not implemented")
        else:
            iso.get_file_from_iso(corpus_dir / rel_path, **{path_arg: abs_path})
    iso.close()
    completed_detector.touch()
    return corpus_dir


FIX_TYPOS = {
    "hm": "hmm",
    "ohh": "oh",
    "hallo": "hello",
    "viscinity": "vicinity",
}

FOREIGN_PATTERN = re.compile(r"<FL>\s*</FL>")
OFF_TALK_PATTERN = re.compile(r"<OT>(.*?)</OT>")
INTERRUPTED_PATTERN1 = re.compile(r"=(\w+)")
INTERRUPTED_PATTERN2 = re.compile(r"(\w+)=")
WHITESPACE_PATTERN = re.compile(r"  +")


def text_normalize(
    text: str,
    silence_sym: str,
    breath_sym: str,
    foreign_sym: str,
    partial_sym: str,  #  When None, will output partial words
    unknown_sym: str,
):

    text = OFF_TALK_PATTERN.sub(r"\1", text)

    result = []
    for w in text.split():
        if w[0] == "@" or w[0] == "~":
            result.append(w[1:])
        elif w in FIX_TYPOS:
            result.append(FIX_TYPOS[w])
        else:
            result.append(w)
    text = " ".join(result).upper()

    text = text.replace("[EMPTY]", silence_sym)
    text = text.replace("[HNOISE]", breath_sym)
    text = FOREIGN_PATTERN.sub(foreign_sym, text)

    if partial_sym == None:
        text = text.replace("=", "")
    else:
        text = INTERRUPTED_PATTERN1.sub(partial_sym, text)
        text = INTERRUPTED_PATTERN2.sub(partial_sym, text)

    for unk in ("[FRAGMENT]", "[NONSENSE]", "[UNKNOWN]"):
        text = text.replace(unk, unknown_sym)

    text = text.replace("AIR SPACE", "AIRSPACE")

    text = WHITESPACE_PATTERN.sub(" ", text)
    text = text.strip()

    return text


def fix_duration(duration: Seconds, sampling_rate: int) -> Seconds:
    """
    A handful of supervision durations do not compute to a round number of
    samples at the original recording sampling rate.

    This causes problem later using compute_num_frames(). Full description:
    https://github.com/lhotse-speech/lhotse/issues/1064

    Return: duration that computes to a round number of samples.
    """
    return compute_num_samples(duration, sampling_rate) / sampling_rate


def prepare_atcosim(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    silence_sym: Optional[str] = "",
    breath_sym: Optional[str] = "",
    foreign_sym: Optional[str] = "<unk>",
    partial_sym: Optional[str] = "<unk>",
    unknown_sym: Optional[str] = "<unk>",
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions

    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param silence_sym: str, silence symbol
    :param breath_sym: str, breath symbol
    :param foreign_sym: str, foreign symbol.
    :param partial_sym: str, partial symbol. When set to None, will output partial words
    :param unknown_sym: str, unknown symbol
    :return: The RecordingSet and SupervisionSet with the keys 'audio' and 'supervisions'.
    """
    if not is_module_available("pandas"):
        raise ImportError("Please 'pip install pandas' first.")
    import pandas as pd

    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    csv_path = corpus_dir / "TXTdata/fulldata.csv"
    df = pd.read_csv(csv_path, quoting=csv.QUOTE_NONE)

    with RecordingSet.open_writer(
        output_dir / f"atcosim_recordings_all.jsonl.gz",
        overwrite=True,
    ) as recs_writer, SupervisionSet.open_writer(
        output_dir / f"atcosim_supervisions_all.jsonl.gz",
        overwrite=True,
    ) as sups_writer:
        for idx, row in tqdm(
            df.iterrows(),
            desc="Preparing",
            total=len(df),
        ):
            if row.recording_corrupt:
                continue

            text = text_normalize(
                row.transcription,
                silence_sym=silence_sym,
                breath_sym=breath_sym,
                foreign_sym=foreign_sym,
                partial_sym=partial_sym,
                unknown_sym=unknown_sym,
            )

            if text == "":
                continue

            wav_path = (
                str(
                    corpus_dir
                    / "WAVdata"
                    / row.directory
                    / row.subdirectory
                    / row.filename
                )
                + ".wav"
            )
            recording = Recording.from_file(wav_path, recording_id=row.recording_id)
            length100 = int(row.length_sec * 100)
            segment = SupervisionSegment(
                id=f"atcosim_{row.filename}_{0:06d}_{length100:06d}",
                recording_id=row.recording_id,
                start=0.0,
                duration=fix_duration(row.length_sec, recording.sampling_rate),
                channel=0,
                language="English",
                text=text,
                speaker=row.speaker_id,
                gender=row.speaker_id[1].upper(),
                custom={"orig_text": row.transcription},
            )

            recs_writer.write(recording)
            sups_writer.write(segment)

    recordings = RecordingSet.from_jsonl_lazy(recs_writer.path)
    supervisions = SupervisionSet.from_jsonl_lazy(sups_writer.path)

    logging.warning(
        "Manifests are lazily materialized. You may want to call `lhotse.qa.fix_manifests()`"
        " to ensure that all supervisions fall within the corresponding recordings."
    )
    return recordings, supervisions



import codecs
import json
from io import BytesIO
from typing import List, Literal, Optional, Union

import lilcom
import numpy as np

from lhotse import Features
from lhotse.array import Array, TemporalArray
from lhotse.shar.utils import to_shar_placeholder
from lhotse.shar.writers.tar import TarWriter


class ArrayTarWriter:
    """
    ArrayTarWriter writes numpy arrays or PyTorch tensors into a tar archive
    that is automatically sharded.

    For floating point tensors, we support the option to use `lilcom` compression.
    Note that `lilcom` is only suitable for log-space features such as log-Mel filter banks.

    Example::

        >>> with ArrayTarWriter("some_dir/fbank.%06d.tar", shard_size=100, compression="lilcom") as w:
        ...     w.write("fbank1", fbank1_array)
        ...     w.write("fbank2", fbank2_array)  # etc.

    It would create files such as ``some_dir/fbank.000000.tar``, ``some_dir/fbank.000001.tar``, etc.

    It's also possible to use ``ArrayTarWriter`` with automatic sharding disabled::

        >>> with ArrayTarWriter("some_dir/fbank.tar", shard_size=None, compression="numpy") as w:
        ...     w.write("fbank1", fbank1_array)
        ...     w.write("fbank2", fbank2_array)  # etc.

    See also: :class:`~lhotse.shar.writers.tar.TarWriter`, :class:`~lhotse.shar.writers.audio.AudioTarWriter`
    """

    def __init__(
        self,
        pattern: str,
        shard_size: Optional[int] = 1000,
        compression: Literal["numpy", "lilcom"] = "numpy",
        lilcom_tick_power: int = -5,
    ):
        self.compression = compression
        self.tar_writer = TarWriter(pattern, shard_size)
        self.lilcom_tick_power = lilcom_tick_power

    def __enter__(self):
        self.tar_writer.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.tar_writer.close()

    @property
    def output_paths(self) -> List[str]:
        return self.tar_writer.output_paths

    def write_placeholder(self, key: str) -> None:
        self.tar_writer.write(f"{key}.nodata", BytesIO())
        self.tar_writer.write(f"{key}.nometa", BytesIO(), count=False)

    def write(
        self,
        key: str,
        value: np.ndarray,
        manifest: Union[Features, Array, TemporalArray],
    ) -> None:

        # Write binary data
        if self.compression == "lilcom":
            assert np.issubdtype(
                value.dtype, np.floating
            ), "Lilcom compression supports only floating-point arrays."
            data = lilcom.compress(value, tick_power=self.lilcom_tick_power)
            stream = BytesIO(data)
            ext = ".llc"
        else:
            stream = BytesIO()
            np.save(stream, value, allow_pickle=False)
            ext = ".npy"

        self.tar_writer.write(key + ext, stream)

        # Write text manifest afterwards
        manifest = to_shar_placeholder(manifest)
        json_stream = BytesIO()
        print(
            json.dumps(manifest.to_dict()),
            file=codecs.getwriter("utf-8")(json_stream),
        )
        json_stream.seek(0)
        self.tar_writer.write(f"{key}.json", json_stream, count=False)



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.this_american_life import (
    download_this_american_life,
    prepare_this_american_life,
)
from lhotse.utils import Pathlike

__all__ = ["this_american_life"]


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option(
    "-f",
    "--force-download",
    is_flag=True,
    default=False,
)
def this_american_life(target_dir: Pathlike, force_download: bool = False):
    """This American Life dataset download."""
    download_this_american_life(target_dir)


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def this_american_life(corpus_dir: Pathlike, output_dir: Pathlike):
    """This American Life data preparation."""
    prepare_this_american_life(corpus_dir, output_dir=output_dir)



"""
About the XBMU-AMDO31 corpus
XBMU-AMDO31 is an open-source Amdo Tibetan speech corpus published by Northwest Minzu University.
publicly available on https://huggingface.co/datasets/syzym/xbmu_amdo31

XBMU-AMDO31 dataset is a speech recognition corpus of Amdo Tibetan dialect.
The open source corpus contains 31 hours of speech data and resources related
to build speech recognition systems,including transcribed texts and a Tibetan
pronunciation lexicon.
(The lexicon is a Tibetan lexicon of the Lhasa dialect, which has been reused
for the Amdo dialect because of the uniformity of the Tibetan language)
The dataset can be used to train a model for Amdo Tibetan Automatic Speech Recognition (ASR).
It was recorded by 66 native speakers of Amdo Tibetan, and the recorded audio was processed and manually inspected.
The dataset has three splits: train, evaluation (dev) and test.Each speaker had approximately 450 sentences,
with a small number of individuals having fewer than 200 sen.

Subset	Hours	Male	Female	Remarks
Train   25.41   27      27      18539 sentences recorded by 54 speakers
Dev     2.81    2       4       2050 sentences recorded by 6 speakers
Test    2.85    3       3       2041 sentences recorded by 6 speakers

Licensing Information
This dataset is distributed under CC BY-SA 4.0.
"""

import logging
import os
import shutil
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available, safe_extract


def download_xbmu_amdo31(
    target_dir: Pathlike = ".",
) -> Path:
    """
    Downdload and untar the dataset
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :return: the path to downloaded and extracted directory with data.
    """
    url = f"https://huggingface.co/datasets/syzym/xbmu_amdo31"
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    corpus_dir = target_dir / "xbmu_amdo31"
    wav_dir = corpus_dir / "data" / "wav"
    train_tar_name = "train.tar.gz"
    dev_tar_name = "dev.tar.gz"
    test_tar_name = "test.tar.gz"

    if not corpus_dir.is_file():
        if is_module_available("git"):
            from git import Repo
        else:
            raise ImportError(
                "In order to download the xbmu-amdo31 corpus from huggingface, please install the relevant dependencies: pip install gitpython"
            )

        logging.info("Start downloading the xbmu-amdo31 corpus")
        try:
            Repo.clone_from(url, corpus_dir)
        except Exception as error:
            print(error)
            raise
        logging.info("Done")

    for tar_name in [train_tar_name, dev_tar_name, test_tar_name]:
        tar_path = wav_dir / tar_name
        extracted_dir = wav_dir / tar_name[:-7]
        completed_detector = extracted_dir / ".completed"
        if completed_detector.is_file():
            logging.info(f"Skipping tar of because {completed_detector} exists.")
            continue
        shutil.rmtree(extracted_dir, ignore_errors=True)
        with tarfile.open(tar_path) as tar:
            safe_extract(tar, path=wav_dir)
        completed_detector.touch()

    return corpus_dir


def prepare_xbmu_amdo31(
    corpus_dir: Pathlike, output_dir: Optional[Pathlike] = None
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    transcript_path = corpus_dir / "data/transcript/transcript_clean.txt"
    transcript_dict = {}
    with open(transcript_path, "r", encoding="utf-8") as f:
        for line in f.readlines():
            idx_transcript = line.split()
            content = " ".join(idx_transcript[1:])
            transcript_dict[idx_transcript[0]] = content
    manifests = defaultdict(dict)
    dataset_parts = ["train", "dev", "test"]
    for part in tqdm(
        dataset_parts,
        desc="Process xbmu_amdo31 audio.",
    ):
        logging.info(f"Processing xbmu_amdo31 subset: {part}")
        # Generate a mapping: utt_id -> (audio_path, audio_info, speaker, text)
        recordings = []
        supervisions = []
        wav_path = corpus_dir / "data" / "wav" / f"{part}"
        count = 0
        for audio_path in wav_path.rglob("**/*.wav"):
            idx = audio_path.stem.split("-")[1]
            speaker = audio_path.parts[-2]
            if idx not in transcript_dict:
                logging.warning(f"No transcript: {idx}")
                logging.warning(f"{audio_path} has no transcript.")
                continue
            text = transcript_dict[idx]
            if not audio_path.is_file():
                logging.warning(f"No such file: {audio_path}")
                continue
            recording = Recording.from_file(audio_path)
            recordings.append(recording)
            count += 1
            segment = SupervisionSegment(
                id=str(count) + "_" + idx,
                recording_id=speaker + "-" + idx,
                start=0.0,
                duration=recording.duration,
                channel=0,
                language="tibetan",
                speaker=speaker,
                text=text.strip(),
            )
            supervisions.append(segment)

        recording_set = RecordingSet.from_recordings(recordings)
        supervision_set = SupervisionSet.from_segments(supervisions)

        # Fix manifests and validate
        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"xbmu_amdo31_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(
                output_dir / f"xbmu_amdo31_recordings_{part}.jsonl.gz"
            )

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



"""
DiPCo is a speech data corpus that simulates a “dinner party” scenario taking place in
an everyday home environment. The corpus was created by recording multiple groups of
four Amazon employee volunteers having a natural conversation in English around a dining
table. The participants were recorded by a single-channel close-talk microphone and
by five far-field 7-microphone array devices positioned at different locations in the
recording room. The dataset contains the audio recordings and human labeled transcripts
of a total of 10 sessions with a duration between 15 and 45 minutes. The corpus was
created to advance in the field of noise robust and distant speech processing and is
intended to serve as a public research and benchmarking data set.

The corpus is made availabe under the CDLA-Permissive license.

More details and download link: https://www.amazon.science/publications/dipco-dinner-party-corpus
"""

import tarfile
from collections import defaultdict
from datetime import datetime as dt
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.recipes.utils import normalize_text_chime6
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, add_durations, resumable_download, safe_extract

CORPUS_URL = "https://s3.amazonaws.com/dipco/DiPCo.tgz"

DATASET_PARTS = {
    "dev": ["S02", "S04", "S05", "S09", "S10"],
    "eval": ["S01", "S03", "S06", "S07", "S08"],
}


def download_dipco(
    target_dir: Pathlike = ".",
    force_download: Optional[bool] = False,
) -> Path:
    """
    Download and untar the dataset
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :return: the path to downloaded and extracted directory with data.
    """
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    tar_path = target_dir / "DiPCo.tgz"
    resumable_download(CORPUS_URL, filename=tar_path, force_download=force_download)
    with tarfile.open(tar_path) as tar:
        safe_extract(tar, path=target_dir)

    return target_dir


def get_session_id(session: str, use_chime7_offset: bool = False) -> str:
    """
    Returns the session ID
    :param session: str, the session ID.
    :param use_chime7_offset: bool, if True, offset session IDs (from CHiME-7 challenge).
    :return: str, the session ID.
    """
    # CHiME-7 challenge offset DiPCo sessions by 24 since the first 24 sessions are
    # used for CHiME-6 sessions.
    if use_chime7_offset:
        session_number = int(session[1:])
        return f"S{24+session_number:02d}"
    else:
        return session


def prepare_dipco(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    mic: Optional[str] = "mdm",
    normalize_text: Optional[str] = "kaldi",
    use_chime7_offset: Optional[bool] = False,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param mic: str, the microphone type to use, choose from "ihm" (close-talk) or "mdm"
        (multi-microphone array) settings. For MDM, there are 5 array devices with 7
        channels each, so the resulting recordings will have 35 channels.
    :param normalize_text: str, the text normalization to apply. Choose from "none",
        "upper", or "kaldi". "kaldi" is the default and is the same normalization
        used in Kaldi's CHiME-6 recipe.
    :param use_chime7_offset: bool, if True, offset session IDs (from CHiME-7 challenge).
    :return: a Dict whose key is the dataset part ("dev" and "eval"), and the value is
        Dicts with the keys 'recordings' and 'supervisions'.
    """
    import json

    import soundfile as sf

    assert mic in ["ihm", "mdm"], "mic must be one of 'ihm' or 'mdm'"

    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    manifests = defaultdict(dict)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    for part in ["dev", "eval"]:
        recordings = []
        supervisions = []

        # First we create the recordings
        if mic == "ihm":
            global_spk_channel_map = {}
            for session in DATASET_PARTS[part]:
                audio_paths = [
                    p for p in (corpus_dir / "audio" / part).rglob(f"{session}_P*.wav")
                ]

                sources = []
                for idx, audio_path in enumerate(audio_paths):
                    sources.append(
                        AudioSource(type="file", channels=[idx], source=str(audio_path))
                    )
                    spk_id = audio_path.stem.split("_")[1]
                    global_spk_channel_map[(session, spk_id)] = idx

                audio_sf = sf.SoundFile(str(audio_paths[0]))

                recordings.append(
                    Recording(
                        id=get_session_id(session, use_chime7_offset),
                        sources=sources,
                        sampling_rate=int(audio_sf.samplerate),
                        num_samples=audio_sf.frames,
                        duration=audio_sf.frames / audio_sf.samplerate,
                    )
                )

        else:
            for session in DATASET_PARTS[part]:
                audio_paths = [
                    p for p in (corpus_dir / "audio" / part).rglob(f"{session}_U*.wav")
                ]

                sources = []
                for idx, audio_path in enumerate(sorted(audio_paths)):
                    sources.append(
                        AudioSource(type="file", channels=[idx], source=str(audio_path))
                    )

                audio_sf = sf.SoundFile(str(audio_paths[0]))

                recordings.append(
                    Recording(
                        id=get_session_id(session, use_chime7_offset),
                        sources=sources,
                        sampling_rate=int(audio_sf.samplerate),
                        num_samples=audio_sf.frames,
                        duration=audio_sf.frames / audio_sf.samplerate,
                    )
                )

        _get_time = lambda x: (
            dt.strptime(x, "%H:%M:%S.%f") - dt(1900, 1, 1)
        ).total_seconds()

        # Then we create the supervisions
        for session in DATASET_PARTS[part]:
            with open(corpus_dir / "transcriptions" / part / f"{session}.json") as f:
                transcript = json.load(f)
                for idx, segment in enumerate(transcript):
                    spk_id = segment["speaker_id"]
                    channel = (
                        global_spk_channel_map[(session, spk_id)]
                        if mic == "ihm"
                        else list(range(35))
                    )
                    start = _get_time(segment["start_time"]["close-talk"])
                    end = _get_time(segment["end_time"]["close-talk"])
                    session_id = get_session_id(session, use_chime7_offset)
                    supervisions.append(
                        SupervisionSegment(
                            id=f"{session_id}-{idx}",
                            recording_id=session_id,
                            start=start,
                            duration=add_durations(end, -start, sampling_rate=16000),
                            channel=channel,
                            text=normalize_text_chime6(
                                segment["words"], normalize=normalize_text
                            ),
                            language="English",
                            speaker=spk_id,
                            gender=segment["gender"],
                            custom={
                                "nativeness": segment["nativeness"],
                                "mother_tongue": segment["mother_tongue"],
                            },
                        )
                    )

        recording_set, supervision_set = fix_manifests(
            RecordingSet.from_recordings(recordings),
            SupervisionSet.from_segments(supervisions),
        )
        # Fix manifests
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"dipco-{mic}_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(
                output_dir / f"dipco-{mic}_recordings_{part}.jsonl.gz"
            )

        manifests[part] = {
            "recordings": recording_set,
            "supervisions": supervision_set,
        }

    return manifests



from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import torch

from lhotse.features.base import FeatureExtractor, register_extractor
from lhotse.utils import Seconds, compute_num_frames_from_samples, is_module_available


@dataclass
class S3PRLSSLConfig:
    """
    In general, the output feature dimension of base model (e.g., wav2vec2) and
    large model (e.g., wav2vec2_large_ll60k) are 768 and 1024, repectively. The
    frame shift (stride) is 0.02s (20ms).

    Please check
        https://github.com/s3prl/s3prl/blob/main/s3prl/upstream/README.md and
        https://s3prl.github.io/s3prl/tutorial/upstream_collection.html
    for details of available self-supervised models.
    """

    sampling_rate: int = 16000
    ssl_model: str = "wav2vec2_large_ll60k"
    layer: int = -1
    frame_shift: float = 0.02
    feature_dim: int = 1024
    device: str = "cpu"

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "S3PRLSSLConfig":
        return S3PRLSSLConfig(**data)


@register_extractor
class S3PRLSSL(FeatureExtractor):
    name = "s3prl-ssl"
    config_type = S3PRLSSLConfig

    def __init__(self, config: Optional[Any] = None):
        super().__init__(config)
        assert is_module_available(
            "s3prl.hub"
        ), "To use s3prl ssl extractors, please install s3prl first."
        import s3prl.hub as hub

        assert self.config.ssl_model in dir(
            hub
        ), f"S3PRL dose not suport model: {self.config.ssl_model}."
        assert (
            self.config.sampling_rate == 16000
        ), f"All the upstream models in S3PRL now only support 16 kHz audio."

        ssl_model = getattr(hub, self.config.ssl_model)()
        self.ssl_model = ssl_model.to(self.config.device)

    @property
    def frame_shift(self) -> Seconds:
        return self.config.frame_shift

    @property
    def sampling_rate(self) -> int:
        return self.config.sampling_rate

    def feature_dim(self, sampling_rate: int) -> int:
        assert (
            sampling_rate == 16000
        ), f"All the upstream models in S3PRL now only support 16 kHz audio."
        return self.config.feature_dim

    def fix_off_by_one_error(self, feats: np.ndarray, num_samples: int) -> np.ndarray:
        # The last frame is usually shorter than the others.
        # We pad it with zeros to make it the same length as others.
        num_frames, num_features = feats.shape
        expected_num_frames = compute_num_frames_from_samples(
            num_samples=num_samples,
            frame_shift=self.frame_shift,
            sampling_rate=self.sampling_rate,
        )
        num_frames_diff = abs(expected_num_frames - num_frames)
        assert num_frames_diff <= 1
        if num_frames_diff == 1:
            pad = np.zeros([1, num_features])
            feats = np.concatenate([feats, pad], axis=0)
        return feats

    def extract_batch(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
        lengths: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
        # S3PRL expects a list of 1D torch tensors.
        if lengths is not None:
            samples = [x[:l] for x, l in zip(samples, lengths)]
        return self.extract(samples=samples, sampling_rate=sampling_rate)

    def extract(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
    ) -> Union[np.ndarray, List[np.ndarray]]:
        assert (
            sampling_rate == 16000
        ), f"All the upstream models in S3PRL now only support 16 kHz audio."

        # s3prl expects a batch of 1D torch tensors.
        # Regardless of input type, we return a numpy array or list of numpy arrays.

        input_is_list = isinstance(samples, list)
        if input_is_list:
            samples = [s.squeeze() for s in samples]
        else:
            samples = samples.squeeze()

        # Convert input to a list of 1D torch tensors.
        if input_is_list or samples.ndim > 1:
            samples = [
                torch.from_numpy(s) if isinstance(s, np.ndarray) else s for s in samples
            ]
        else:
            # The user passed a single array/tensor of shape (num_samples,)
            samples = [
                torch.from_numpy(samples)
                if isinstance(samples, np.ndarray)
                else samples
            ]

        samples = [s.to(self.config.device) for s in samples]
        lengths = [s.shape[0] for s in samples]

        self.ssl_model.eval()
        with torch.no_grad():
            feats = self.ssl_model(samples)["hidden_states"][self.config.layer]
        feats = feats.squeeze()

        if feats.ndim == 2:
            # The user passed a single array/tensor of shape (num_samples,)
            feats = feats.cpu().numpy()
            feats = self.fix_off_by_one_error(feats, lengths[0])
            if input_is_list:
                feats = [feats]
        else:
            # The user passed a batch of arrays/tensors of shape (num_samples,)
            # Convert the padded sequence to a list of 1D torch tensors.
            out_lens = [
                compute_num_frames_from_samples(
                    num_samples, self.config.frame_shift, self.config.sampling_rate
                )
                for num_samples in lengths
            ]
            feats = [f[:l].cpu().numpy() for f, l in zip(feats, out_lens)]
            feats = [self.fix_off_by_one_error(f, l) for f, l in zip(feats, lengths)]

            # If all items are of the same shape, stack them into a single array.
            if all(item.shape == feats[0].shape for item in feats[1:]):
                feats = np.stack(feats, axis=0)

        return feats



import logging
import random
from pathlib import Path
from typing import Callable, Dict, Generator, Iterable, Optional, Sequence, Tuple, Union

import torch
from cytoolz import compose_left

from lhotse import CutSet, Seconds
from lhotse.cut.set import deserialize_cut
from lhotse.dataset.dataloading import get_rank, get_world_size
from lhotse.dataset.sampling.base import SamplingDiagnostics
from lhotse.lazy import Dillable
from lhotse.serialization import decode_json_line
from lhotse.utils import Pathlike

PathlikeAndScale = Tuple[Pathlike, float]


class StatelessSampler(torch.utils.data.Sampler, Dillable):
    """
    An infinite and stateless cut sampler that selects data at random from one or more cut manifests.
    The main idea is to make training resumption easy while guaranteeing the data seen each
    time by the model is shuffled differently.
    It discards the notion of an "epoch" and it never finishes iteration.
    It makes no strong guarantees about avoiding data duplication, but in practice you would
    rarely see duplicated data.

    The recommended way to use this sampler is by placing it into a dataloader worker
    subprocess with Lhotse's :class:``~lhotse.dataset.iterable_dataset.IterableDatasetWrapper``,
    so that each worker has its own sampler replica that uses a slightly different random seed::

        >>> import torch
        >>> import lhotse
        >>> dloader = torch.utils.data.DataLoader(
        ...     lhotse.dataset.iterable_dataset.IterableDatasetWrapper(
        ...         dataset=lhotse.dataset.K2SpeechRecognitionDataset(...),
        ...         sampler=StatelessSampler(...),
        ...     ),
        ...     batch_size=None,
        ...     num_workers=4,
        ... )

    This sampler's design was originally proposed by Dan Povey. For details see:
    https://github.com/lhotse-speech/lhotse/issues/1096

    Example 1: Get a non-bucketing :class:``.StatelessSampler``::

        >>> sampler = StatelessSampler(
        ...     cuts_paths=["data/cuts_a.jsonl", "data/cuts_b.jsonl"],
        ...     index_path="data/files.idx",
        ...     max_duration=600.0,
        ... )

    Example 2: Get a bucketing :class:``.StatelessSampler``::

        >>> sampler = StatelessSampler(
        ...     cuts_paths=["data/cuts_a.jsonl", "data/cuts_b.jsonl"],
        ...     index_path="data/files.idx",
        ...     max_duration=600.0,
        ...     num_buckets=50,
        ...     quadratic_duration=30.0,
        ... )

    Example 3: Get a bucketing :class:``.StatelessSampler`` with scaled weights for each cutset::

        >>> sampler = StatelessSampler(
        ...     cuts_paths=[
        ...         ("data/cuts_a.jsonl", 2.0),
        ...         ("data/cuts_b.jsonl", 1.0),
        ...     ],
        ...     index_path="data/files.idx",
        ...     max_duration=600.0,
        ...     num_buckets=50,
        ...     quadratic_duration=30.0,
        ... )

    .. note:: This sampler works only with uncompressed jsonl manifests, as it creates extra index files
     with line byte offsets to quickly find and sample JSON lines.
     This means this sampler will not work with Webdataset and Lhotse Shar data format.

    :param cuts_paths: Path, or list of paths, or list of tuples of (path, scale) to cutset files.
    :param index_path: Path to a file that contains the index of all cutsets and their line count
        (will be auto-created the first time this object is initialized).
    :param base_seed: Int, user-provided part of the seed used to initialize the RNG
        for sampling (each node and worker are still going to produce different results).
        When continuing the training it should be a function of the number of training steps
        to ensure the model doesn't see identical mini-batches again.
    :param max_duration: Maximum total number of audio seconds in a mini-batch (dynamic batch size).
    :param max_cuts: Maximum number of examples in a mini-batch (static batch size).
    :param num_buckets: If set, enables bucketing (each mini-batch has examples of a similar duration).
    :param quadratic_duration: If set, adds a penalty term for longer duration cuts.
        Works well with models that have quadratic time complexity to keep GPU utilization similar
        when using bucketing. Suggested values are between 30 and 45.
    """

    def __init__(
        self,
        cuts_paths: Union[Pathlike, Iterable[Pathlike], Iterable[PathlikeAndScale]],
        index_path: Pathlike,
        base_seed: int,
        max_duration: Optional[Seconds] = None,
        max_cuts: Optional[int] = None,
        num_buckets: Optional[int] = None,
        quadratic_duration: Optional[Seconds] = None,
    ) -> None:
        super().__init__(data_source=None)

        # Parse file paths and scales
        self.paths = []
        self.scales = []
        if isinstance(cuts_paths, (Path, str)):
            # Single path input
            self.paths.append(Path(cuts_paths))
            self.scales.append(1.0)
        else:
            # Multiple path input
            cuts_paths = list(cuts_paths)
            self.paths = []
            if isinstance(cuts_paths[0], (Path, str)):
                # Without scales
                for p in cuts_paths:
                    assert isinstance(
                        p, (Path, str)
                    ), "Mixing paths with and without scales is not allowed."
                    self.paths.append(Path(p))
                    self.scales.append(1.0)
            else:
                for tpl in cuts_paths:
                    # With scales
                    assert len(tpl) == 2, (
                        f"Expected (path, scale) but got: {tpl} "
                        f"[note: mixing paths with and without scales is not allowed]"
                    )
                    p, scale = tpl
                    assert isinstance(
                        p, (Path, str)
                    ), f"Path must be a string or Path, got: {p}"
                    assert isinstance(
                        scale, (int, float)
                    ), f"Scale must be an int or float, got: {scale}"
                    self.paths.append(Path(p))
                    self.scales.append(scale)

        self.index_path = index_path
        self.max_duration = max_duration
        self.max_cuts = max_cuts
        self.num_buckets = num_buckets
        self.quadratic_duration = quadratic_duration
        self.base_seed = base_seed
        assert any(
            v is not None for v in (self.max_duration, self.max_cuts)
        ), "At least one of max_duration or max_cuts has to be set."

        self.diagnostics = SamplingDiagnostics()
        self.index = ManifestIndex(self.paths, self.index_path)
        self.scaled_line_counts = [
            lc * scale
            for lc, scale in zip(self.index.line_counts.values(), self.scales)
        ]
        self._transforms = []
        # DDP related info
        self.rank = get_rank()
        self.world_size = get_world_size()

    def map(self, fn: Callable[[CutSet], CutSet]) -> "StatelessSampler":
        """Apply ``fn`` to each mini-batch of ``CutSet`` before yielding it."""
        self._transforms.append(fn)
        return self

    def state_dict(self) -> Dict:
        """Stub state_dict method that returns nothing - this sampler is stateless."""
        return {}

    def load_state_dict(self, state_dict: Dict) -> None:
        """Stub load_state_dict method that does nothing - this sampler is stateless."""
        return

    def __iter__(self) -> Generator[CutSet, None, None]:
        from lhotse.dataset import DynamicBucketingSampler, DynamicCutSampler

        worker_info = torch.utils.data.get_worker_info()
        worker_id = 0 if worker_info is None else worker_info.id
        my_id = worker_id + 1000 * self.rank
        seed = self.base_seed + my_id
        rng = random.Random(seed)
        logging.info(
            f"[{type(self).__name__}] Initialized sampler RNG with seed {seed} (== base_seed={self.base_seed} + my_id={my_id}) "
            f"[ddp_rank={self.rank} worker_id={worker_id}]"
        )

        def _inner():
            """
            Infinite generator of cuts.
            Each cut is samples in two steps:
            - first we select a cutset file, weighted by line count (num cuts)
            - then we randomly select a line from that file using uniform distribution
            """
            n = 0
            while True:
                # Choose a file
                path = rng.choices(self.paths, self.scaled_line_counts)[0]
                # Choose a line
                line_offsets = self.index.line_offsets[path]
                begin_idx = rng.randrange(len(line_offsets) - 1)
                begin, end = line_offsets[begin_idx], line_offsets[begin_idx + 1]
                # Read JSON line and initialize a Cut object
                with path.open() as f:
                    f.seek(begin)
                    line = f.read(end - begin)
                data = decode_json_line(line)
                cut = deserialize_cut(data)
                # Update cut ID since the same item may land in a single mini-batch
                # (note: CutSet prohibits two items sharing the same ID)
                cut.id = f"{cut.id}_it{n}"
                yield cut
                n += 1

        if self.num_buckets is not None and self.num_buckets > 1:
            inner_sampler = DynamicBucketingSampler(
                _inner(),
                max_duration=self.max_duration,
                max_cuts=self.max_cuts,
                num_buckets=self.num_buckets,
                shuffle=False,
                drop_last=False,
                quadratic_duration=self.quadratic_duration,
                world_size=1,
                rank=0,
            )
        else:
            inner_sampler = DynamicCutSampler(
                _inner(),
                max_duration=self.max_duration,
                max_cuts=self.max_cuts,
                shuffle=False,
                drop_last=False,
                world_size=1,
                rank=0,
            )
        inner_sampler.map(compose_left(*self._transforms))
        self.diagnostics = inner_sampler.diagnostics
        yield from inner_sampler

    def get_report(self) -> str:
        """Returns a string describing the statistics of the sampling process so far."""
        return self.diagnostics.get_report()


class ManifestIndex:
    """
    An index of line count and line offset for each cutset manifest.
    When created for the first time, it writes a .jsonl.idx file for each .jsonl file that contains byte offsets for each line.
    It also writes a file at ``index_path`` that has the line count and path for each manifest.
    When this object is instantiated again (e.g. when resuming training), it will just load the contents of existing files from disk.

    Objects of this class expose two members: ``line_counts: Dict[Path, int]`` and ``line_offsets: Dict[Path, List[int]]`` to simplify working with manifests.

    :param manifest_paths: A list of paths to cut sets.
    :param index_path: A path where we should write the line count index (if it doesn't exist).
    :param force: When true, we'll ignore existing files and reindex the cutsets.
    """

    def __init__(
        self,
        manifest_paths: Sequence[Pathlike],
        index_path: Pathlike,
        force: bool = False,
    ) -> None:
        self.line_counts: Dict[Path, int] = {}
        self.line_offsets: Dict[Path, Tuple[int]] = {}
        for p in map(Path, manifest_paths):
            assert (
                p.suffix == ".jsonl"
            ), f"We only support uncompressed .jsonl files in this sampler, but received: {p}"

            offset_path = p.with_suffix(".jsonl.idx")
            if offset_path.is_file() and not force:
                offsets = self._load(offset_path)
            else:
                offsets = self._process(p, offset_path)
            self.line_counts[p] = len(offsets)
            self.line_offsets[p] = offsets

        # Write a single cutset index in format:
        # <number-of-lines> <cutset-path>
        # Example:
        # 10015 data/cuts-part-0001.jsonl
        # 376101 data/cuts-part-0002.jsonl
        # 572 data/cuts-part-0003.jsonl
        if not index_path.is_file() or force:
            with index_path.open("w") as index_f:
                for p, lc in self.line_counts.items():
                    print(f"{lc} {p}", file=index_f)

    def _load(self, file_index: Path) -> Tuple[int]:
        with file_index.open() as f:
            offsets = tuple(map(int, f))
        return offsets

    def _process(self, manifest: Path, file_index: Path) -> Tuple[int]:
        # Write line index for each cutset in format <begin-byte> per line, e.g.:
        # 0
        # 214
        # 357
        # ...
        offsets = [0]
        with manifest.open() as cuts_f, file_index.open("w") as index_f:
            print(0, file=index_f)
            line = cuts_f.readline()
            while line:
                offsets.append(cuts_f.tell())
                print(offsets[-1], file=index_f)
                line = cuts_f.readline()
        return tuple(offsets)



from .backend import (
    audio_backend,
    available_audio_backends,
    get_current_audio_backend,
    get_default_audio_backend,
    get_ffmpeg_torchaudio_info_enabled,
    info,
    read_audio,
    save_audio,
    set_current_audio_backend,
    set_ffmpeg_torchaudio_info_enabled,
)
from .recording import Recording
from .recording_set import RecordingSet
from .source import AudioSource
from .utils import (
    AudioLoadingError,
    DurationMismatchError,
    VideoInfo,
    get_audio_duration_mismatch_tolerance,
    null_result_on_audio_loading_error,
    set_audio_duration_mismatch_tolerance,
    suppress_audio_loading_errors,
)

__all__ = [
    "AudioSource",
    "Recording",
    "RecordingSet",
    "AudioLoadingError",
    "DurationMismatchError",
    "VideoInfo",
    "audio_backend",
    "available_audio_backends",
    "get_current_audio_backend",
    "get_default_audio_backend",
    "get_audio_duration_mismatch_tolerance",
    "get_ffmpeg_torchaudio_info_enabled",
    "info",
    "read_audio",
    "save_audio",
    "set_current_audio_backend",
    "set_audio_duration_mismatch_tolerance",
    "set_ffmpeg_torchaudio_info_enabled",
    "null_result_on_audio_loading_error",
    "suppress_audio_loading_errors",
]



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.daily_talk import download_daily_talk, prepare_daily_talk
from lhotse.utils import Pathlike


@prepare.command()
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option("--num-jobs", type=int, default=1, help="Number of parallel workers.")
def daily_talk(corpus_dir: Pathlike, output_dir: Pathlike, num_jobs: int):
    """
    DailyTalk recording and supervision manifest preparation.
    """
    prepare_daily_talk(corpus_dir, output_dir, num_jobs=num_jobs)


@download.command()
@click.argument("target_dir", type=click.Path())
@click.option("--force-download", is_flag=True, help="Force download.")
def daily_talk(target_dir: Pathlike, force_download: bool = False):
    """
    Download DailyTalk dataset.
    """
    download_daily_talk(target_dir, force_download)



"""
The ICMC-ASR Grand Challenge dataset is collected in a hybrid electric vehicle with speakers sitting in different positions, including the driver seat and passenger seats. The total number of speakers is over 160 and all of them are native Chinese speakers speaking Mandarin without strong accents. To comprehensively capture speech signals of the entire cockpit, two types of recording devices are used: far-field and near-field recording devices. 8 distributed microphones are placed at four seats in the car, which are the driver's seat (DS01C01, DX01C01), the passenger seat (DS02C01, DX02C01), the rear right seat (DS03C01, DX03C01) and the rear left seat (DS04C01, DX04C01). Additionally, 2 linear microphone arrays, each consisting of 2 microphones, are placed on the display screen (DL01C01, DL02C02) and at the center of the inner sunroof (DL02C01, DL02C02), respectively. All 12 channels of far-field data are time-synchronized and included in the released dataset as far-field data. For transcription purposes, each speaker wears a high-fidelity headphone to record near-field audio, denoted by the seat where the speaker is situated. Specifically, DA01, DA02, DA03, and DA04 represent the driver seat, passenger seat, rear right seat and rear left seat, respectively. The near-field data only have single-channel audio recordings. Additionally, a sizable real noise dataset is provided, following the recording setup of the far-filed data but without speaker talking, to facilitate research of in-car scenario data simulation technology.

Participants can obtain the datasets at https://icmcasr.org - please download the datasets manually.
"""

import logging
import os
from collections import defaultdict
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.audio.backend import info
from lhotse.qa import fix_manifests, validate_recordings_and_supervisions
from lhotse.recipes.utils import manifests_exist, normalize_text_alimeeting
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available

ICMCASR = (
    "train",
    "dev",
    "eval_track1",
)  # TODO: Support eval2 by people who have access to it.
POSITION = ("DA01", "DA02", "DA03", "DA04")
# ignore "DX05C01", "DX06C01",
# which are 2-channel reference signals for AEC.
# see https://github.com/MrSupW/ICMC-ASR_Baseline/tree/main
SDM_POSITION = ("DX01C01", "DX02C01", "DX03C01", "DX04C01")


def _parse_utterance(
    corpus_dir: Pathlike,
    section_path: Pathlike,
    mic: str,
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    if not is_module_available("textgrid"):
        raise ValueError(
            "To prepare ICMC ASR data, please 'pip install textgrid' first."
        )
    import textgrid

    recordings = []
    segments = []
    for position in POSITION:
        text_path = (section_path / (position + ".TextGrid")).resolve()
        if not text_path.is_file():
            continue
        if mic == "ihm":
            audio_paths = [(section_path / (position + ".wav")).resolve()]
            recording_ids = [
                str(section_path / position)
                .replace(str(corpus_dir) + "/", "")
                .replace("/", "-")
            ]
        elif mic == "sdm":
            audio_paths = [
                (section_path / (sdm_position + ".wav")).resolve()
                for sdm_position in SDM_POSITION
            ]
            recording_ids = [
                str(section_path / sdm_position)
                .replace(str(corpus_dir) + "/", "")
                .replace("/", "-")
                + f"-{position}"
                for sdm_position in SDM_POSITION
            ]
        elif mic == "mdm":
            audio_paths = ["fake_audio_path_for_mdm"]
            recording_ids = [
                str(section_path / "DXmixC01")
                .replace(str(corpus_dir) + "/", "")
                .replace("/", "-")
                + f"-{position}"
            ]
        else:
            raise ValueError(f"Unsupported mic type: {mic}")

        for audio_path, recording_id in zip(audio_paths, recording_ids):
            if mic == "mdm":
                channel_paths = [
                    (section_path / (position + ".wav")).resolve()
                    for position in SDM_POSITION
                ]
                audio_info = info(
                    channel_paths[0],
                    force_opus_sampling_rate=None,
                    force_read_audio=False,
                )
                recordings.append(
                    Recording(
                        id=recording_id,
                        sources=[
                            AudioSource(
                                type="file",
                                channels=[idx],
                                source=str(audio_path),
                            )
                            for idx, audio_path in enumerate(channel_paths)
                        ],
                        sampling_rate=16000,
                        num_samples=audio_info.frames,
                        duration=audio_info.duration,
                    )
                )
            # check if audio_path exists, if not, then skip
            else:
                if not audio_path.is_file():
                    # give some warning
                    logging.warning(
                        f"Audio file {audio_path} does not exist - skipping."
                    )
                    continue
                recordings.append(
                    Recording.from_file(path=audio_path, recording_id=recording_id)
                )

            tg = textgrid.TextGrid.fromFile(str(text_path))
            assert len(tg.tiers) == 1, f"Expected 1 tier, found {len(tg.tiers)} tiers."
            tier = tg.tiers[0]
            speaker = tier.name
            for i, interval in enumerate(tier.intervals):
                if interval.mark != "":
                    start = interval.minTime
                    end = interval.maxTime
                    text = interval.mark
                    segment = SupervisionSegment(
                        id=f"{recording_id}-{round(start * 1000):06}-{round(end * 1000):06}",
                        recording_id=recording_id,
                        start=start,
                        duration=round(end - start, 4),
                        channel=0 if mic in ["sdm", "ihm"] else list(range(4)),
                        language="Chinese",
                        speaker=speaker,
                        text=normalize_text_alimeeting(text),
                    )
                    segments.append(segment)

    return recordings, segments


def _prepare_subset(
    subset: str,
    corpus_dir: Pathlike,
    mic: str,
    num_jobs: int = 1,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param subset: str, the name of the subset.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet for train and valid.
    """
    corpus_dir = Path(corpus_dir)
    part_path = corpus_dir / subset
    sections = os.listdir(part_path)

    with ThreadPoolExecutor(num_jobs) as ex:
        futures = []
        recording_set = []
        supervision_set = []
        for section in tqdm(sections, desc="Distributing tasks"):
            section_path = part_path / section
            futures.append(ex.submit(_parse_utterance, corpus_dir, section_path, mic))

        for future in tqdm(futures, desc="Processing"):
            result = future.result()
            if result is None:
                continue
            recordings, segments = result
            recording_set.extend(recordings)
            supervision_set.extend(segments)

        recording_set = RecordingSet.from_recordings(recording_set)
        supervision_set = SupervisionSet.from_segments(supervision_set)

        # Fix manifests
        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

    return recording_set, supervision_set


def prepare_icmcasr(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    mic: str = "ihm",
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Path to the ICMC-ASR dataset.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    logging.info("Preparing ICMC-ASR...")

    subsets = ICMCASR
    if mic == "ihm":
        subsets = ("train", "dev")

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)

    for part in tqdm(subsets, desc="Dataset parts"):
        logging.info(f"Processing ICMC-ASR subset: {part}")
        if manifests_exist(
            part=part,
            output_dir=output_dir,
            prefix=f"icmcasr-{mic}",
            suffix="jsonl.gz",
        ):
            logging.info(f"ICMC-ASR subset: {part} already prepared - skipping.")
            continue

        recording_set, supervision_set = _prepare_subset(
            part, corpus_dir, mic, num_jobs
        )

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"icmcasr-{mic}_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(
                output_dir / f"icmcasr-{mic}_recordings_{part}.jsonl.gz"
            )

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



import warnings
from dataclasses import dataclass
from decimal import ROUND_HALF_UP
from typing import Dict, Optional, Tuple

import numpy as np
import torch

from lhotse.augmentation.transform import AudioTransform
from lhotse.utils import (
    Seconds,
    compute_num_samples,
    during_docs_build,
    is_module_available,
    is_torchaudio_available,
    perturb_num_samples,
)


@dataclass
class Speed(AudioTransform):
    """
    Speed perturbation effect, the same one as invoked with `sox speed` in the command line.

    It resamples the signal back to the input sampling rate, so the number of output samples will
    be smaller or greater, depending on the speed factor.
    """

    factor: float

    def __call__(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        check_for_torchaudio()
        resampler = get_or_create_resampler(
            round(sampling_rate * self.factor), sampling_rate
        )
        augmented = resampler(torch.from_numpy(samples))
        return augmented.numpy()

    def reverse_timestamps(
        self, offset: Seconds, duration: Optional[Seconds], sampling_rate: int
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method helps estimate the original offset and duration for a recording
        before speed perturbation was applied.
        We need this estimate to know how much audio to actually load from disk during the
        call to ``load_audio()``.
        """
        start_sample = compute_num_samples(offset, sampling_rate)
        num_samples = (
            compute_num_samples(duration, sampling_rate)
            if duration is not None
            else None
        )
        start_sample = perturb_num_samples(start_sample, 1 / self.factor)
        num_samples = (
            perturb_num_samples(num_samples, 1 / self.factor)
            if num_samples is not None
            else None
        )
        return (
            start_sample / sampling_rate,
            num_samples / sampling_rate if num_samples is not None else None,
        )


_precompiled_resamplers: Dict[Tuple[int, int], torch.nn.Module] = {}


def get_or_create_resampler(
    source_sampling_rate: int, target_sampling_rate: int
) -> torch.nn.Module:
    check_for_torchaudio()
    global _precompiled_resamplers

    tpl = (source_sampling_rate, target_sampling_rate)
    if tpl not in _precompiled_resamplers:
        check_torchaudio_version()
        import torchaudio

        _precompiled_resamplers[tpl] = torchaudio.transforms.Resample(
            source_sampling_rate, target_sampling_rate
        )
    return _precompiled_resamplers[tpl]


@dataclass
class Resample(AudioTransform):
    """
    Resampling effect, the same one as invoked with `sox rate` in the command line.
    """

    source_sampling_rate: int
    target_sampling_rate: int

    def __post_init__(self):
        self.source_sampling_rate = int(self.source_sampling_rate)
        self.target_sampling_rate = int(self.target_sampling_rate)
        if not is_torchaudio_available():
            assert is_module_available(
                "scipy"
            ), "In order to use resampling, either torchaudio or scipy needs to be installed."
        else:
            self.resampler = get_or_create_resampler(
                self.source_sampling_rate, self.target_sampling_rate
            )

    def __call__(self, samples: np.ndarray, *args, **kwargs) -> np.ndarray:
        if self.source_sampling_rate == self.target_sampling_rate:
            return samples

        if is_torchaudio_available():
            if isinstance(samples, np.ndarray):
                samples = torch.from_numpy(samples)
            augmented = self.resampler(samples)
            return augmented.numpy()
        else:
            import scipy

            gcd = np.gcd(self.source_sampling_rate, self.target_sampling_rate)
            augmented = scipy.signal.resample_poly(
                samples,
                up=self.target_sampling_rate // gcd,
                down=self.source_sampling_rate // gcd,
                axis=-1,
            )
            return augmented

    def reverse_timestamps(
        self, offset: Seconds, duration: Optional[Seconds], sampling_rate: int
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method helps estimate the original offset and duration for a recording
        before resampling was applied.
        We need this estimate to know how much audio to actually load from disk during the
        call to ``load_audio()``.

        In case of resampling, the timestamps might change slightly when using non-trivial
        pairs of sampling rates, e.g. 16kHz -> 22.05kHz, because the number of samples in
        the resampled audio might actually correspond to incrementally larger/smaller duration.
        E.g. 16kHz, 235636 samples correspond to 14.72725s duration; after resampling to 22.05kHz,
        it is 324736 samples which correspond to 14.727256235827664s duration.
        """
        if self.source_sampling_rate == self.target_sampling_rate:
            return offset, duration

        old_num_samples = compute_num_samples(
            offset, self.source_sampling_rate, rounding=ROUND_HALF_UP
        )
        old_offset = old_num_samples / self.source_sampling_rate
        if duration is not None:
            old_num_samples = compute_num_samples(
                duration, self.source_sampling_rate, rounding=ROUND_HALF_UP
            )
            old_duration = old_num_samples / self.source_sampling_rate
        else:
            old_duration = None
        return old_offset, old_duration


@dataclass
class Tempo(AudioTransform):
    """Tempo perturbation effect, the same one as invoked with `sox tempo` in the command line.

    Compared to speed perturbation, tempo preserves pitch.
    It resamples the signal back to the input sampling rate, so the number of output samples will
    be smaller or greater, depending on the tempo factor.
    """

    factor: float

    def __call__(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        check_for_torchaudio()
        check_torchaudio_version()
        import torchaudio

        sampling_rate = int(sampling_rate)  # paranoia mode
        if isinstance(samples, np.ndarray):
            samples = torch.from_numpy(samples)

        augmented, new_sampling_rate = torchaudio.sox_effects.apply_effects_tensor(
            samples, sampling_rate, [["tempo", str(self.factor)]]
        )
        return augmented.numpy()

    def reverse_timestamps(
        self,
        offset: Seconds,
        duration: Optional[Seconds],
        sampling_rate: int,
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method helps estimate the original offset and duration for a recording
        before tempo perturbation was applied.
        We need this estimate to know how much audio to actually load from disk during the
        call to ``load_audio()``.
        """
        start_sample = compute_num_samples(offset, sampling_rate)
        num_samples = (
            compute_num_samples(duration, sampling_rate)
            if duration is not None
            else None
        )
        start_sample = perturb_num_samples(start_sample, 1 / self.factor)
        num_samples = (
            perturb_num_samples(num_samples, 1 / self.factor)
            if num_samples is not None
            else None
        )
        return (
            start_sample / sampling_rate,
            num_samples / sampling_rate if num_samples is not None else None,
        )


@dataclass
class Volume(AudioTransform):
    """
    Volume perturbation effect, the same one as invoked with `sox vol` in the command line.
    It applies given gain (factor) to the input, without any postprocessing (such as a limiter).
    """

    factor: float

    def __call__(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        # We only support the simplest case in SoX which is multiplication by a gain value.
        # https://github.com/chirlu/sox/blob/42b3557e13e0fe01a83465b672d89faddbe65f49/src/vol.c#L149
        return samples * self.factor

    def reverse_timestamps(
        self,
        offset: Seconds,
        duration: Optional[Seconds],
        sampling_rate: Optional[int],  # Not used, made for compatibility purposes
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method just returnes the original offset and duration as volume perturbation
        doesn't change any these audio properies.
        """
        return offset, duration


def check_torchaudio_version():
    import torchaudio
    from packaging.version import parse as _version

    if not during_docs_build() and _version(torchaudio.__version__) < _version("0.7"):
        warnings.warn(
            "Torchaudio SoX effects chains are only introduced in version 0.7 - "
            "please upgrade your PyTorch to 1.7.1 and torchaudio to 0.7.2 (or higher) "
            "to use them."
        )


def check_for_torchaudio():
    if not is_torchaudio_available():
        raise RuntimeError(
            "This transform is not supported in torchaudio-free Lhotse installation. "
            "Please install torchaudio and try again."
        )



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.voxconverse import download_voxconverse, prepare_voxconverse
from lhotse.utils import Pathlike


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option("--force-download", is_flag=True, default=False, help="Force download")
def voxconverse(target_dir: Pathlike, force_download=False):
    """VoxConverse dataset download."""
    download_voxconverse(target_dir, force_download=force_download)


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "--split-test",
    is_flag=True,
    default=False,
    help="Split test part into dev and test parts",
)
def voxconverse(corpus_dir: Pathlike, output_dir: Pathlike, split_test: bool = False):
    """VoxConverse data preparation."""
    prepare_voxconverse(corpus_dir, output_dir=output_dir, split_test=split_test)



import re
from typing import List, NamedTuple, Sequence

import torch

from lhotse.supervision import AlignmentItem

from .base import FailedToAlign, ForcedAligner


class ASRForcedAligner(ForcedAligner):
    def __init__(
        self, bundle_name: str = "WAV2VEC2_ASR_BASE_960H", device: str = "cpu", **kwargs
    ):
        import torchaudio

        super().__init__(device=device)
        self.bundle_name = bundle_name
        self.bundle = getattr(torchaudio.pipelines, bundle_name)
        self.model = self.bundle.get_model().to(device)
        self.labels = self.bundle.get_labels()
        self.dictionary = {c: i for i, c in enumerate(self.labels)}
        self.discard_symbols = _make_discard_symbols_regex(self.labels)

    @property
    def sample_rate(self) -> int:
        return self.bundle.sample_rate

    def normalize_text(self, text: str, **kwargs) -> str:
        return _normalize_text(text, self.discard_symbols)

    def align(self, audio: torch.Tensor, transcript: str) -> List[AlignmentItem]:
        tokens = [self.dictionary[c] for c in transcript]

        with torch.inference_mode():
            emissions, _ = self.model(audio)
            emissions = torch.log_softmax(emissions, dim=-1)
        emission = emissions[0].cpu()

        trellis = _get_trellis(emission, tokens)
        path = _backtrack(trellis, emission, tokens)

        segments = _merge_repeats(path, transcript)

        word_segments = _merge_words(segments)

        # Ratio of number of samples to number of frames
        ratio = audio.size(1) / emission.size(0)
        return [
            AlignmentItem(
                symbol=ws.label,
                start=round(int(ratio * ws.start) / self.sample_rate, ndigits=8),
                duration=round(
                    int(ratio * (ws.end - ws.start)) / self.sample_rate,
                    ndigits=8,
                ),
                score=ws.score,
            )
            for ws in word_segments
        ]


def _make_discard_symbols_regex(labels: Sequence[str]) -> re.Pattern:
    return re.compile(rf"[^{' '.join(labels)}]")


def _normalize_text(text: str, discard_symbols: re.Pattern) -> str:
    from lhotse.utils import is_module_available

    assert is_module_available(
        "num2words"
    ), "To align with torchaudio, please run 'pip install num2words' for number to word normalization."
    from num2words import num2words

    text = re.sub(r"(\d+)", lambda x: num2words(int(x.group(0))), text)
    return re.sub(discard_symbols, "", text.upper().replace(" ", "|"))


def _get_trellis(
    emission: torch.Tensor, tokens: Sequence[int], blank_id: int = 0
) -> torch.Tensor:
    num_frame = emission.size(0)
    num_tokens = len(tokens)

    # Trellis has extra diemsions for both time axis and tokens.
    # The extra dim for tokens represents <SoS> (start-of-sentence)
    # The extra dim for time axis is for simplification of the code.
    trellis = torch.empty((num_frame + 1, num_tokens + 1))
    trellis[0, 0] = 0
    trellis[1:, 0] = torch.cumsum(emission[:, 0], 0)
    trellis[0, -num_tokens:] = -float("inf")
    trellis[-num_tokens:, 0] = float("inf")

    for t in range(num_frame):
        trellis[t + 1, 1:] = torch.maximum(
            # Score for staying at the same token
            trellis[t, 1:] + emission[t, blank_id],
            # Score for changing to the next token
            trellis[t, :-1] + emission[t, tokens],
        )
    return trellis


class Point(NamedTuple):
    token_index: int
    time_index: int
    score: float


def _backtrack(
    trellis: torch.Tensor,
    emission: torch.Tensor,
    tokens: Sequence[int],
    blank_id: int = 0,
) -> List[Point]:
    # Note:
    # j and t are indices for trellis, which has extra dimensions
    # for time and tokens at the beginning.
    # When referring to time frame index `T` in trellis,
    # the corresponding index in emission is `T-1`.
    # Similarly, when referring to token index `J` in trellis,
    # the corresponding index in transcript is `J-1`.
    j = trellis.size(1) - 1
    t_start = torch.argmax(trellis[:, j]).item()

    path = []
    for t in range(t_start, 0, -1):
        # 1. Figure out if the current position was stay or change
        # Note (again):
        # `emission[J-1]` is the emission at time frame `J` of trellis dimension.
        # Score for token staying the same from time frame J-1 to T.
        stayed = trellis[t - 1, j] + emission[t - 1, blank_id]
        # Score for token changing from C-1 at T-1 to J at T.
        changed = trellis[t - 1, j - 1] + emission[t - 1, tokens[j - 1]]

        # 2. Store the path with frame-wise probability.
        prob = emission[t - 1, tokens[j - 1] if changed > stayed else 0].exp().item()
        # Return token index and time index in non-trellis coordinate.
        path.append(Point(j - 1, t - 1, prob))

        # 3. Update the token
        if changed > stayed:
            j -= 1
            if j == 0:
                break
    else:
        raise FailedToAlign()
    return path[::-1]


class Segment(NamedTuple):
    label: str
    start: int
    end: int
    score: float

    def __repr__(self):
        return f"{self.label}\t({self.score:4.2f}): [{self.start:5d}, {self.end:5d})"

    @property
    def length(self):
        return self.end - self.start


def _merge_repeats(path: List[Point], transcript: str) -> List[Segment]:
    i1, i2 = 0, 0
    segments = []
    while i1 < len(path):
        while i2 < len(path) and path[i1].token_index == path[i2].token_index:
            i2 += 1
        score = sum(path[k].score for k in range(i1, i2)) / (i2 - i1)
        segments.append(
            Segment(
                transcript[path[i1].token_index],
                path[i1].time_index,
                path[i2 - 1].time_index + 1,
                score,
            )
        )
        i1 = i2
    return segments


def _merge_words(segments: List[Segment], separator: str = "|") -> List[Segment]:
    words = []
    i1, i2 = 0, 0
    while i1 < len(segments):
        if i2 >= len(segments) or segments[i2].label == separator:
            if i1 != i2:
                segs = segments[i1:i2]
                word = "".join([seg.label for seg in segs])
                score = sum(seg.score * seg.length for seg in segs) / sum(
                    seg.length for seg in segs
                )
                words.append(
                    Segment(word, segments[i1].start, segments[i2 - 1].end, score)
                )
            i1 = i2 + 1
            i2 = i1
        else:
            i2 += 1
    return words



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.atcosim import download_atcosim, prepare_atcosim
from lhotse.utils import Pathlike

__all__ = ["atcosim"]


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
def atcosim(target_dir: Pathlike):
    """ATCOSIM download."""
    download_atcosim(target_dir)


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option("--silence-sym", type=str, default="")
@click.option("--breath-sym", type=str, default="")
@click.option("--foreign-sym", type=str, default="<unk>")
@click.option("--partial-sym", type=str, default="<unk>")
@click.option("--unknown-sym", type=str, default="<unk>")
def atcosim(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    silence_sym: str,
    breath_sym: str,
    foreign_sym: str,
    partial_sym: str,
    unknown_sym: str,
):
    """ATCOSIM data preparation."""
    prepare_atcosim(
        corpus_dir,
        output_dir=output_dir,
        silence_sym=silence_sym,
        breath_sym=breath_sym,
        foreign_sym=foreign_sym,
        partial_sym=partial_sym,
        unknown_sym=unknown_sym,
    )



from typing import Any, Dict

import torch

from lhotse import CutSet
from lhotse.dataset.collation import collate_video


class UnsupervisedAudioVideoDataset(torch.utils.data.Dataset):
    """
    A basic dataset that loads, pads, collates, and returns video and audio tensors.

    Returns:

    .. code-block::

        {
            'video': (B x NumFrames x Color x Height x Width) uint8 tensor
            'video_lens': (B, ) int32 tensor
            'audio': (B x NumChannels x NumSamples) float32 tensor
            'audio_lens': (B, ) int32 tensor
            'cuts': CutSet of length B
        }
    """

    def __getitem__(self, cuts: CutSet) -> Dict[str, Any]:
        video, video_lens, audio, audio_lens, cuts = collate_video(
            cuts, fault_tolerant=True
        )
        return {
            "cuts": cuts,
            "video": video,
            "video_lens": video_lens,
            "audio": audio,
            "audio_lens": audio_lens,
        }



import logging
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import torch

from lhotse.audio import Recording
from lhotse.audio.utils import VideoInfo
from lhotse.cut.base import Cut
from lhotse.features import FeatureExtractor
from lhotse.supervision import SupervisionSegment
from lhotse.utils import (
    LOG_EPSILON,
    Pathlike,
    Seconds,
    compute_num_frames,
    compute_num_samples,
    fastcopy,
    perturb_num_samples,
    uuid4,
)


@dataclass
class PaddingCut(Cut):
    """
    :class:`~lhotse.cut.PaddingCut` is a dummy :class:`~lhotse.cut.Cut` that doesn't refer to
    actual recordings or features --it simply returns zero samples in the time domain
    and a specified features value in the feature domain.
    Its main role is to be appended to other cuts to make them evenly sized.

    Please refer to the documentation of :class:`~lhotse.cut.Cut` to learn more about using cuts.

    See also:

        - :class:`lhotse.cut.Cut`
        - :class:`lhotse.cut.MonoCut`
        - :class:`lhotse.cut.MixedCut`
        - :class:`lhotse.cut.CutSet`
    """

    id: str
    duration: Seconds
    sampling_rate: int
    feat_value: float

    # For frequency domain
    num_frames: Optional[int] = None
    num_features: Optional[int] = None
    frame_shift: Optional[float] = None

    # For time domain
    num_samples: Optional[int] = None
    video: Optional[VideoInfo] = None

    # Dict for storing padding values for custom array attributes
    custom: Optional[dict] = None

    @property
    def start(self) -> Seconds:
        return 0

    @property
    def supervisions(self):
        return []

    @property
    def channel(self) -> int:
        return 0

    @property
    def has_features(self) -> bool:
        return self.num_frames is not None

    @property
    def has_recording(self) -> bool:
        return self.num_samples is not None

    @property
    def has_video(self) -> bool:
        return self.has_recording and self.video is not None

    @property
    def num_channels(self) -> int:
        return 1

    def has(self, field: str) -> bool:
        if field == "recording":
            return self.has_recording
        elif field == "features":
            return self.has_features
        elif field == "video":
            return self.has_video
        else:
            return self.custom is not None and field in self.custom

    @property
    def recording_id(self) -> str:
        return "PAD"

    # noinspection PyUnusedLocal
    def load_features(self, *args, **kwargs) -> Optional[np.ndarray]:
        if self.has_features:
            return (
                np.ones((self.num_frames, self.num_features), np.float32)
                * self.feat_value
            )
        return None

    # noinspection PyUnusedLocal
    def load_audio(self, *args, **kwargs) -> Optional[np.ndarray]:
        if self.has_recording:
            return np.zeros(
                (1, compute_num_samples(self.duration, self.sampling_rate)), np.float32
            )
        return None

    def load_video(
        self,
        with_audio: bool = True,
    ) -> Optional[Tuple[torch.Tensor, Optional[torch.Tensor]]]:
        if self.has_video:
            audio = None
            if with_audio:
                audio = torch.zeros(
                    1,
                    compute_num_samples(self.duration, self.sampling_rate),
                    dtype=torch.float32,
                )
            return (
                torch.zeros(
                    self.video.num_frames,
                    3,
                    self.video.height,
                    self.video.width,
                    dtype=torch.uint8,
                ),
                audio,
            )
        return None

    # noinspection PyUnusedLocal
    def truncate(
        self,
        *,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        keep_excessive_supervisions: bool = True,
        preserve_id: bool = False,
        **kwargs,
    ) -> "PaddingCut":
        new_duration = self.duration - offset if duration is None else duration
        assert new_duration > 0.0
        return fastcopy(
            self,
            id=self.id if preserve_id else str(uuid4()),
            duration=new_duration,
            feat_value=self.feat_value,
            num_frames=compute_num_frames(
                duration=new_duration,
                frame_shift=self.frame_shift,
                sampling_rate=self.sampling_rate,
            )
            if self.num_frames is not None
            else None,
            num_samples=compute_num_samples(
                duration=new_duration, sampling_rate=self.sampling_rate
            )
            if self.num_samples is not None
            else None,
        )

    # noinspection PyUnusedLocal
    def extend_by(
        self,
        *,
        duration: Seconds,
        direction: str = "both",
        preserve_id: bool = False,
        pad_silence: bool = True,
    ) -> "PaddingCut":
        """
        Return a new PaddingCut with region extended by the specified duration.

        :param duration: The duration by which to extend the cut.
        :param direction: string, 'left', 'right' or 'both'. Determines whether the cut should
            be extended to the left, right or both sides. By default, the cut is extended by
            the specified duration on both sides.
        :param preserve_id: When ``True``, preserves the cut ID from before padding.
            Otherwise, generates a new random ID (default).
        :param pad_silence: See usage in :func:`lhotse.cut.MonoCut.extend_by`. It is ignored here.
        :return: an extended PaddingCut.
        """
        new_duration = self.duration + duration
        if direction == "both":
            new_duration += duration
        assert new_duration > 0.0
        return fastcopy(
            self,
            id=self.id if preserve_id else str(uuid4()),
            duration=new_duration,
            feat_value=self.feat_value,
            num_frames=compute_num_frames(
                duration=new_duration,
                frame_shift=self.frame_shift,
                sampling_rate=self.sampling_rate,
            )
            if self.num_frames is not None
            else None,
            num_samples=compute_num_samples(
                duration=new_duration, sampling_rate=self.sampling_rate
            )
            if self.num_samples is not None
            else None,
        )

    def pad(
        self,
        duration: Seconds = None,
        num_frames: int = None,
        num_samples: int = None,
        pad_feat_value: float = LOG_EPSILON,
        direction: str = "right",
        preserve_id: bool = False,
        pad_value_dict: Optional[Dict[str, Union[int, float]]] = None,
    ) -> Cut:
        """
        Return a new MixedCut, padded with zeros in the recording, and ``pad_feat_value`` in each feature bin.

        The user can choose to pad either to a specific `duration`; a specific number of frames `max_frames`;
        or a specific number of samples `num_samples`. The three arguments are mutually exclusive.

        :param duration: The cut's minimal duration after padding.
        :param num_frames: The cut's total number of frames after padding.
        :param num_samples: The cut's total number of samples after padding.
        :param pad_feat_value: A float value that's used for padding the features.
            By default we assume a log-energy floor of approx. -23 (1e-10 after exp).
        :param direction: string, 'left', 'right' or 'both'. Determines whether the padding is added before or after
            the cut.
        :param preserve_id: When ``True``, preserves the cut ID from before padding.
            Otherwise, generates a new random ID (default).
        :param pad_value_dict: Optional dict that specifies what value should be used
            for padding arrays in custom attributes.
        :return: a padded MixedCut if duration is greater than this cut's duration, otherwise ``self``.
        """
        from .set import pad

        return pad(
            self,
            duration=duration,
            num_frames=num_frames,
            num_samples=num_samples,
            pad_feat_value=pad_feat_value,
            direction=direction,
            preserve_id=preserve_id,
            pad_value_dict=pad_value_dict,
        )

    def resample(self, sampling_rate: int, affix_id: bool = False) -> "PaddingCut":
        """
        Return a new ``MonoCut`` that will lazily resample the audio while reading it.
        This operation will drop the feature manifest, if attached.
        It does not affect the supervision.

        :param sampling_rate: The new sampling rate.
        :param affix_id: Should we modify the ID (useful if both versions of the same
            cut are going to be present in a single manifest).
        :return: a modified copy of the current ``MonoCut``.
        """
        assert self.has_recording, "Cannot resample a MonoCut without Recording."
        return fastcopy(
            self,
            id=f"{self.id}_rs{sampling_rate}" if affix_id else self.id,
            sampling_rate=sampling_rate,
            num_samples=compute_num_samples(self.duration, sampling_rate),
            num_frames=None,
            num_features=None,
            frame_shift=None,
        )

    def perturb_speed(self, factor: float, affix_id: bool = True) -> "PaddingCut":
        """
        Return a new ``PaddingCut`` that will "mimic" the effect of speed perturbation
        on ``duration`` and ``num_samples``.

        :param factor: The speed will be adjusted this many times (e.g. factor=1.1 means 1.1x faster).
        :param affix_id: When true, we will modify the ``PaddingCut.id`` field
            by affixing it with "_sp{factor}".
        :return: a modified copy of the current ``PaddingCut``.
        """
        # Pre-conditions
        if self.has_features:
            logging.warning(
                "Attempting to perturb speed on a MonoCut that references pre-computed features. "
                "The feature manifest will be detached, as we do not support feature-domain "
                "speed perturbation."
            )
            new_num_frames = None
            new_num_features = None
            new_frame_shift = None
        else:
            new_num_frames = self.num_frames
            new_num_features = self.num_features
            new_frame_shift = self.frame_shift
        new_num_samples = perturb_num_samples(self.num_samples, factor)
        new_duration = new_num_samples / self.sampling_rate
        return fastcopy(
            self,
            id=f"{self.id}_sp{factor}" if affix_id else self.id,
            num_samples=new_num_samples,
            duration=new_duration,
            num_frames=new_num_frames,
            num_features=new_num_features,
            frame_shift=new_frame_shift,
        )

    def perturb_tempo(self, factor: float, affix_id: bool = True) -> "PaddingCut":
        """
        Return a new ``PaddingCut`` that will "mimic" the effect of tempo perturbation
        on ``duration`` and ``num_samples``.

        Compared to speed perturbation, tempo preserves pitch.
        :param factor: The tempo will be adjusted this many times (e.g. factor=1.1 means 1.1x faster).
        :param affix_id: When true, we will modify the ``PaddingCut.id`` field
            by affixing it with "_tp{factor}".
        :return: a modified copy of the current ``PaddingCut``.
        """
        # Pre-conditions
        if self.has_features:
            logging.warning(
                "Attempting to perturb tempo on a MonoCut that references pre-computed features. "
                "The feature manifest will be detached, as we do not support feature-domain "
                "tempo perturbation."
            )
            new_num_frames = None
            new_num_features = None
            new_frame_shift = None
        else:
            new_num_frames = self.num_frames
            new_num_features = self.num_features
            new_frame_shift = self.frame_shift
        new_num_samples = perturb_num_samples(self.num_samples, factor)
        new_duration = new_num_samples / self.sampling_rate
        return fastcopy(
            self,
            id=f"{self.id}_tp{factor}" if affix_id else self.id,
            num_samples=new_num_samples,
            duration=new_duration,
            num_frames=new_num_frames,
            num_features=new_num_features,
            frame_shift=new_frame_shift,
        )

    def perturb_volume(self, factor: float, affix_id: bool = True) -> "PaddingCut":
        """
        Return a new ``PaddingCut`` that will "mimic" the effect of volume perturbation
        on amplitude of samples.

        :param factor: The volume will be adjusted this many times (e.g. factor=1.1 means 1.1x louder).
        :param affix_id: When true, we will modify the ``PaddingCut.id`` field
            by affixing it with "_vp{factor}".
        :return: a modified copy of the current ``PaddingCut``.
        """

        return fastcopy(self, id=f"{self.id}_vp{factor}" if affix_id else self.id)

    def reverb_rir(
        self,
        rir_recording: Optional["Recording"] = None,
        normalize_output: bool = True,
        early_only: bool = False,
        affix_id: bool = True,
        rir_channels: List[int] = [0],
        room_rng_seed: Optional[int] = None,
        source_rng_seed: Optional[int] = None,
    ) -> "PaddingCut":
        """
        Return a new ``PaddingCut`` that will "mimic" the effect of reverberation with impulse response
        on original samples.

        :param rir_recording: The impulse response to use for convolving.
        :param normalize_output: When true, output will be normalized to have energy as input.
        :param early_only: When true, only the early reflections (first 50 ms) will be used.
        :param affix_id: When true, we will modify the ``PaddingCut.id`` field
            by affixing it with "_rvb".
        :param rir_channels: The channels of the impulse response to use.
        :return: a modified copy of the current ``PaddingCut``.
        """

        return fastcopy(self, id=f"{self.id}_rvb" if affix_id else self.id)

    def normalize_loudness(
        self, target: float, affix_id: bool = False, **kwargs
    ) -> "PaddingCut":
        """
        Return a new ``PaddingCut`` that will "mimic" the effect of loudness normalization

        :param target: The target loudness in dBFS.
        :param affix_id: When true, we will modify the ``DataCut.id`` field
            by affixing it with "_ln{target}".
        :return: a modified copy of the current ``DataCut``.
        """
        return fastcopy(
            self,
            id=f"{self.id}_ln{target}" if affix_id else self.id,
        )

    def drop_features(self) -> "PaddingCut":
        """Return a copy of the current :class:`.PaddingCut`, detached from ``features``."""
        assert (
            self.has_recording
        ), f"Cannot detach features from a MonoCut with no Recording (cut ID = {self.id})."
        return fastcopy(self, num_frames=None, num_features=None, frame_shift=None)

    def drop_recording(self) -> "PaddingCut":
        """Return a copy of the current :class:`.PaddingCut`, detached from ``recording``."""
        assert (
            self.has_features
        ), f"Cannot detach recording from a PaddingCut with no Features (cut ID = {self.id})."
        return fastcopy(self, num_samples=None)

    def drop_supervisions(self) -> "PaddingCut":
        """Return a copy of the current :class:`.PaddingCut`, detached from ``supervisions``."""
        return self

    def drop_alignments(self) -> "PaddingCut":
        """Return a copy of the current :class:`.PaddingCut`, detached from ``alignments``."""
        return self

    def compute_and_store_features(
        self, extractor: FeatureExtractor, *args, **kwargs
    ) -> Cut:
        """
        Returns a new PaddingCut with updates information about the feature dimension and number of
        feature frames, depending on the ``extractor`` properties.
        """
        return fastcopy(
            self,
            num_features=extractor.feature_dim(self.sampling_rate),
            num_frames=compute_num_frames(
                duration=self.duration,
                frame_shift=extractor.frame_shift,
                sampling_rate=self.sampling_rate,
            ),
            frame_shift=extractor.frame_shift,
        )

    def fill_supervision(self, *args, **kwargs) -> "PaddingCut":
        """
        Just for consistency with :class`.MonoCut` and :class:`.MixedCut`.
        """
        return self

    def move_to_memory(self, *args, **kwargs) -> "PaddingCut":
        """
        Just for consistency with :class`.MonoCut` and :class:`.MixedCut`.
        """
        return self

    def map_supervisions(self, transform_fn: Callable[[Any], Any]) -> "PaddingCut":
        """
        Just for consistency with :class:`.MonoCut` and :class:`.MixedCut`.

        :param transform_fn: a dummy function that would be never called actually.
        :return: the PaddingCut itself.
        """
        return self

    def merge_supervisions(self, *args, **kwargs) -> "PaddingCut":
        """
        Just for consistency with :class:`.MonoCut` and :class:`.MixedCut`.

        :return: the PaddingCut itself.
        """
        return self

    def filter_supervisions(
        self, predicate: Callable[[SupervisionSegment], bool]
    ) -> "PaddingCut":
        """
        Just for consistency with :class:`.MonoCut` and :class:`.MixedCut`.

        :param predicate: A callable that accepts `SupervisionSegment` and returns bool
        :return: a modified MonoCut
        """
        return self

    @staticmethod
    def from_dict(data: dict) -> "PaddingCut":
        # Remove "type" field if exists
        data.pop("type", None)
        return PaddingCut(**data)

    def with_features_path_prefix(self, path: Pathlike) -> "PaddingCut":
        return self

    def with_recording_path_prefix(self, path: Pathlike) -> "PaddingCut":
        return self



import logging

import click

from lhotse.bin.modes import prepare
from lhotse.recipes.must_c import prepare_must_c
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option("--tgt-lang", type=str, help="The target language, e.g., zh, de, fr.")
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def must_c(corpus_dir: Pathlike, output_dir: Pathlike, tgt_lang, num_jobs: int):
    """MUST-C speech translation data preparation."""
    logging.basicConfig(level=logging.INFO)
    prepare_must_c(
        corpus_dir,
        output_dir=output_dir,
        tgt_lang=tgt_lang,
        num_jobs=num_jobs,
    )



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes import download_grid, prepare_grid
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "--with-supervisions/--no-supervisions",
    default=True,
    help="Note: using supervisions might discard some recordings that do not have them.",
)
@click.option("-j", "--jobs", default=1, type=int, help="The number of parallel jobs.")
def grid(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    with_supervisions: bool,
    jobs: int,
):
    """Grid audio-visual speech corpus preparation."""
    prepare_grid(
        corpus_dir,
        output_dir=output_dir,
        with_supervisions=with_supervisions,
        num_jobs=jobs,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
def grid(
    target_dir: Pathlike,
):
    """Grid audio-visual speech corpus download."""
    download_grid(target_dir)



import random
from concurrent.futures import ProcessPoolExecutor, as_completed
from functools import partial
from pathlib import Path
from typing import List, Optional

import click
import tqdm

from lhotse import CutSet, Fbank, FeatureExtractor
from lhotse.bin.modes.cli_base import cli
from lhotse.features.io import MemoryRawWriter
from lhotse.shar import ArrayTarWriter
from lhotse.utils import Pathlike


@cli.group()
def shar():
    """Lhotse Shar format for optimized I/O commands"""
    pass


@shar.command(context_settings=dict(show_default=True))
@click.argument("cuts", type=click.Path(exists=True, dir_okay=False))
@click.argument("outdir", type=click.Path())
@click.option(
    "-a",
    "--audio",
    default="none",
    type=click.Choice(["none", "wav", "flac", "mp3", "opus"]),
    help="Format in which to export audio (disabled by default, enabling will make a copy of the data)",
)
@click.option(
    "-f",
    "--features",
    default="none",
    type=click.Choice(["none", "lilcom", "numpy"]),
    help="Format in which to export features (disabled by default, enabling will make a copy of the data)",
)
@click.option(
    "-c",
    "--custom",
    multiple=True,
    default=[],
    help="Custom fields to export. Use syntax NAME:FORMAT, e.g.: -c target_recording:flac -c embedding:numpy. Use format options for audio and features depending on the custom fields type, or 'jsonl' for metadata.",
)
@click.option(
    "-s",
    "--shard-size",
    type=int,
    default=1000,
    help="The number of cuts in a single shard.",
)
@click.option(
    "--shuffle/--no-shuffle",
    default=True,
    help="Should we shuffle the cuts before splitting into shards.",
)
@click.option(
    "--fault-tolerant/--fast-fail",
    default=False,
    help="Should we skip over cuts that failed to load data or raise an error.",
)
@click.option("--seed", default=0, type=int, help="Random seed.")
@click.option(
    "-j",
    "--num-jobs",
    default=1,
    type=int,
    help="Number of parallel workers. We recommend to keep this number low on machines "
    "with slow disks as the speed of I/O will likely be the bottleneck.",
)
@click.option("-v", "--verbose", count=True)
def export(
    cuts: str,
    outdir: str,
    audio: str,
    features: str,
    custom: List[str],
    shard_size: int,
    shuffle: bool,
    fault_tolerant: bool,
    seed: int,
    num_jobs: int,
    verbose: bool,
):
    """
    Export CutSet from CUTS into Lhotse Shar format in OUTDIR.

    This script partitions the input manifest into smaller pieces called shards
    with SHARD_SIZE cuts per shard. The input is optionally shuffled.
    In addition to sharding, the user can choose to export AUDIO or FEATURES
    into sequentially readable tar files with a selected compression type.
    This typically yields very high speedups vs random read formats such as HDF5,
    especially on slower disks or clusters, at the expense of a data copy.

    The result is readable in Python using: CutSet.from_shar(OUTDIR)
    """
    cuts: CutSet = CutSet.from_file(cuts)

    if shuffle:
        cuts = cuts.shuffle(rng=random.Random(seed))

    fields = {}
    if audio != "none":
        fields["recording"] = audio
    if features != "none":
        fields["features"] = features
    if custom:
        for item in custom:
            key, fmt = item.split(":")
            fields[key] = fmt

    Path(outdir).mkdir(parents=True, exist_ok=True)
    cuts.to_shar(
        output_dir=outdir,
        fields=fields,
        shard_size=shard_size,
        num_jobs=num_jobs,
        fault_tolerant=fault_tolerant,
        verbose=verbose,
    )


@shar.command(context_settings=dict(show_default=True))
@click.argument("shar_dir", type=click.Path(exists=True, file_okay=False))
@click.option(
    "-f",
    "--feature-config",
    type=click.Path(exists=True, dir_okay=False),
    help="Optional manifest specifying feature extractor configuration (use Fbank by default).",
)
@click.option(
    "-c",
    "--compression",
    type=click.Choice(["lilcom", "numpy"]),
    default="lilcom",
    help="Which compression to use (lilcom is lossy, numpy is lossless).",
)
@click.option(
    "-j", "--num-jobs", default=1, type=int, help="Number of parallel workers."
)
@click.option("-v", "--verbose", count=True)
def compute_features(
    shar_dir: str,
    feature_config: Optional[str],
    compression: str,
    num_jobs: int,
    verbose: int,
):
    """
    Compute features for Lhotse Shar cuts stored in SHAR_DIR.

    The features are computed sequentially on CPU within shards,
    and parallelized across shards up to NUM_JOBS concurrent workers.

    FEATURE_CONFIG defines the feature extractor type and settings.
    You can generate default feature extractor settings with:
    lhotse feat write-default-config --help
    """
    shards = [
        {
            "cuts": [p],
            "recording": [p.with_name("".join(["recording", p.suffixes[0], ".tar"]))],
        }
        for p in Path(shar_dir).glob("cuts.*.jsonl*")
    ]
    progbar = lambda x: x
    if verbose:
        click.echo(f"Computing features for {len(shards)} shards.")
        progbar = partial(tqdm.tqdm, desc="Shard progress", total=len(shards))

    futures = []
    with ProcessPoolExecutor(num_jobs) as ex:
        for shard in shards:
            cuts_path = shard["cuts"][0]
            shard_idx = cuts_path.name.split(".")[1]
            output_path = cuts_path.with_name(f"features.{shard_idx}.tar")
            futures.append(
                ex.submit(
                    compute_features_one_shard,
                    cuts=CutSet.from_shar(shard),
                    feature_config=feature_config,
                    output_path=output_path,
                    compression=compression,
                )
            )
        for f in progbar(as_completed(futures)):
            f.result()


def compute_features_one_shard(
    cuts: CutSet, feature_config: Pathlike, output_path: Pathlike, compression: str
):
    extractor = (
        FeatureExtractor.from_yaml(feature_config)
        if feature_config is not None
        else Fbank()
    )
    in_memory = MemoryRawWriter()
    with ArrayTarWriter(
        output_path, shard_size=None, compression=compression
    ) as writer:
        for cut in cuts:
            cut = cut.compute_and_store_features(extractor, in_memory)
            writer.write(key=cut.id, value=cut.load_features(), manifest=cut.features)



from .common import AugmentFn
from .loudness import LoudnessNormalization
from .rir import ReverbWithImpulseResponse
from .torchaudio import *
from .transform import AudioTransform
from .utils import FastRandomRIRGenerator, convolve1d
from .wpe import DereverbWPE, dereverb_wpe_numpy, dereverb_wpe_torch



import logging
from collections import defaultdict
from functools import partial
from typing import Any, List, Optional, Union

import numpy as np
from tqdm import tqdm

from lhotse import RecordingSet, SupervisionSet, dill_enabled
from lhotse.cut import CutSet, MixedCut, MixTrack
from lhotse.cut.set import mix
from lhotse.parallel import parallel_map
from lhotse.utils import add_durations, uuid4
from lhotse.workflows.meeting_simulation.base import (
    MAX_TASKS_WAITING,
    BaseMeetingSimulator,
    MeetingSampler,
    reverberate_cuts,
)


class ConversationalMeetingSimulator(BaseMeetingSimulator):
    """
    This simulator uses the method described in https://arxiv.org/abs/2204.00890 and
    implemented in https://github.com/BUTSpeechFIT/EEND_dataprep. Note that a similar
    method of meeting simulation is also implemented in
    https://github.com/jsalt2020-asrdiar/jsalt2020_simulate.

    The basic idea is to sample the silence and overlap durations collectively for all
    speakers so that we get similar speech/silence/overlap characteristics as the
    original distribution. In general, this method produces more realistic overlap
    durations than the `SpeakerIndependentMeetingSimulator`. This is done by learning
    the histogram of 3 distributions: same speaker pause, different speaker pause, and
    different speaker overlap. In this implementation, we learn the histograms from
    provided data, otherwise we use the initialization values as shift of a Gamma
    distribution with scale 1.0.
    """

    def __init__(
        self,
        same_spk_pause: float = 1.0,
        diff_spk_pause: float = 1.0,
        diff_spk_overlap: float = 2.0,
        prob_diff_spk_overlap: float = 0.5,
    ):
        """
        :param same_spk_pause: the mean pause duration between utterances of the same
            speaker. [Default: 2.0]
        :param diff_spk_pause: the mean pause duration between utterances of different
            speakers. [Default: 3.0]
        :param diff_spk_overlap: the mean overlap duration between utterances of
            different speakers. [Default: 2.0]
        :param prob_diff_spk_overlap: the probability of overlap between utterances of
            different speakers. [Default: 0.5]
        """
        super().__init__()
        for duration in [same_spk_pause, diff_spk_pause, diff_spk_overlap]:
            assert duration is None or duration > 0, "Durations must be > 0."

        self.same_spk_pause = same_spk_pause
        self.diff_spk_pause = diff_spk_pause
        self.diff_spk_overlap = diff_spk_overlap
        self.prob_diff_spk_overlap = prob_diff_spk_overlap

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__} "
            f"(same_spk_pause={self.same_spk_pause:.2f}, "
            f"diff_spk_pause={self.diff_spk_pause:.2f}, "
            f"diff_spk_overlap={self.diff_spk_overlap:.2f}, "
            f"prob_diff_spk_overlap={self.prob_diff_spk_overlap:.2f})"
        )

    def _init_defaults(self):
        from scipy.stats import gamma

        self.same_spk_pause_dist = gamma(a=1.0, scale=1.0, loc=self.same_spk_pause)
        self.diff_spk_pause_dist = gamma(a=1.0, scale=1.0, loc=self.diff_spk_pause)
        self.diff_spk_overlap_dist = gamma(a=1.0, scale=1.0, loc=self.diff_spk_overlap)

    def _compute_histogram_dist(self, values: np.ndarray) -> Any:
        """
        Compute the histogram of the given values and return the bin edges and the
        corresponding probabilities.
        """
        from scipy.stats import rv_histogram

        hist, bin_edges = np.histogram(values, bins=100, density=True)
        return rv_histogram((hist, bin_edges))

    @dill_enabled(True)
    def fit(self, meetings: Optional[SupervisionSet] = None) -> None:
        """
        Learn the distribution of the meeting parameters from a given dataset.
        :param meetings: a SupervisionSet containing the meetings to be used for
        """
        if meetings is None:
            logging.info(f"No meetings provided, using default parameters.")
            self._init_defaults()
            return

        assert isinstance(
            meetings, SupervisionSet
        ), "The meetings must be provided as a SupervisionSet."

        from cytoolz.itertoolz import groupby

        same_spk_pause_values = []
        diff_spk_pause_values = []
        diff_spk_overlap_values = []

        recording_segments = groupby(
            lambda s: s.recording_id,
            sorted(meetings, key=lambda s: (s.recording_id, s.start)),
        )

        for segments in recording_segments.values():
            for i in range(1, len(segments)):
                if segments[i].speaker == segments[i - 1].speaker:
                    same_spk_pause_values.append(
                        segments[i].start - segments[i - 1].end
                    )
                    continue
                if segments[i].start > segments[i - 1].end:
                    diff_spk_pause_values.append(
                        segments[i].start - segments[i - 1].end
                    )
                else:
                    diff_spk_overlap_values.append(
                        segments[i - 1].end - segments[i].start
                    )

        # Generate histogram distributions.

        self.same_spk_pause_dist = self._compute_histogram_dist(
            np.array(same_spk_pause_values)
        )
        self.diff_spk_pause_dist = self._compute_histogram_dist(
            np.array(diff_spk_pause_values)
        )
        self.diff_spk_overlap_dist = self._compute_histogram_dist(
            np.array(diff_spk_overlap_values)
        )
        self.prob_diff_spk_overlap = (
            len(diff_spk_overlap_values)
            / (len(diff_spk_pause_values) + len(diff_spk_overlap_values))
            if (len(diff_spk_pause_values) + len(diff_spk_overlap_values)) > 0
            else 0.5
        )

        # Update the parameters.
        self.same_spk_pause = self.same_spk_pause_dist.mean()
        self.diff_spk_pause = self.diff_spk_pause_dist.mean()
        self.diff_spk_overlap = self.diff_spk_overlap_dist.mean()

        print(f"Learned parameters: {self}")

    def _create_mixture(
        self, utterances: CutSet, allow_3fold_overlap: bool = False
    ) -> MixedCut:
        """
        Create a MixedCut object from a list of MonoCuts (utterances).
        We sample pauses and/or overlaps from the initilized or learned distributions.
        Then, we create a MixedCut where each track represents a different speaker.

        :param utterances: a CutSet containing the utterances to be mixed.
        :param allow_3fold_overlap: if True, allow 3-fold overlaps between speakers.
            [Default: False]
        :return: a MixedCut object.
        """
        speakers = utterances.speakers

        # Generate pause/overlap timings for all utterances.
        N = len(utterances)
        same_spk_pauses = [round(x, 2) for x in self.same_spk_pause_dist.rvs(size=N)]
        diff_spk_pauses = [round(x, 2) for x in self.diff_spk_pause_dist.rvs(size=N)]
        diff_spk_overlaps = [
            round(x, 2) for x in self.diff_spk_overlap_dist.rvs(size=N)
        ]
        diff_spk_bernoulli = self.bernoulli.rvs(p=self.prob_diff_spk_overlap, size=N)

        utterances = list(utterances)
        # First sample offsets for each utterance. These are w.r.t. start of the meeting.
        # For each subsequent utterance, we sample a pause or overlap time from the
        # corresponding distribution. Then, we add the pause/overlap time to the offset
        # of the previous utterance to get the offset of the current utterance.
        offsets = [0.0]
        cur_offset = utterances[0].duration

        # We keep track of the end time of the last utterance for each speaker.
        first_spk = utterances[0].supervisions[0].speaker
        last_utt_end = {spkr: 0.0 for spkr in speakers}
        last_utt_end[first_spk] = cur_offset
        last_utt_end_times = sorted(list(last_utt_end.values()), reverse=True)
        sr = utterances[0].sampling_rate

        for i in range(1, len(utterances)):
            cur_spk = utterances[i].supervisions[0].speaker
            prev_spk = utterances[i - 1].supervisions[0].speaker
            if cur_spk == prev_spk:
                ot = same_spk_pauses[i]
            else:
                if diff_spk_bernoulli[i] == 0:
                    # No overlap between speakers.
                    ot = diff_spk_pauses[i]
                else:
                    # Overlap between speakers.
                    ot = diff_spk_overlaps[i]
                    if len(last_utt_end_times) > 1 and not allow_3fold_overlap:
                        # second term for ensuring same speaker's utterances do not overlap.
                        # third term for ensuring the maximum number of overlaps is two.
                        ot = min(
                            ot,
                            add_durations(
                                cur_offset, -last_utt_end[cur_spk], sampling_rate=sr
                            ),
                            add_durations(
                                cur_offset, -last_utt_end_times[1], sampling_rate=sr
                            ),
                        )
                    else:
                        ot = min(
                            ot,
                            add_durations(
                                cur_offset, -last_utt_end[cur_spk], sampling_rate=sr
                            ),
                        )
                    ot = -ot

            cur_offset = add_durations(cur_offset, ot, sampling_rate=sr)
            offsets.append(cur_offset)
            cur_offset = add_durations(
                cur_offset, utterances[i].duration, sampling_rate=sr
            )

            # Update last_utt_end and last_utt_end_times.
            last_utt_end[cur_spk] = cur_offset
            last_utt_end_times = sorted(list(last_utt_end.values()), reverse=True)
            cur_offset = last_utt_end_times[0]

        # Group utterances (and offsets) by speaker. First, we will sort the utterances
        # by their offsets.
        utterances, offsets = zip(*sorted(zip(utterances, offsets), key=lambda x: x[1]))
        spk_tracks = defaultdict(list)
        for utt, offset in zip(utterances, offsets):
            spk_tracks[utt.supervisions[0].speaker].append((utt, offset))

        tracks = []
        for spk, spk_utts in spk_tracks.items():
            track, start = spk_utts[0]
            for utt, offset in spk_utts[1:]:
                track = mix(
                    track,
                    utt,
                    offset=add_durations(offset, -start, sampling_rate=sr),
                    allow_padding=True,
                )
            track = MixTrack(cut=track, type=type(track), offset=start)
            tracks.append(track)

        # sort tracks by track offset
        tracks = sorted(tracks, key=lambda x: x.offset)
        return MixedCut(id=str(uuid4()), tracks=tracks)

    @dill_enabled(True)
    def simulate(
        self,
        cuts: CutSet,
        num_meetings: Optional[int] = None,
        num_repeats: Optional[int] = None,
        num_speakers_per_meeting: Union[int, List[int]] = 2,
        speaker_count_probs: Optional[List[float]] = None,
        max_duration_per_speaker: Optional[float] = 20.0,
        max_utterances_per_speaker: Optional[int] = 5,
        allow_3fold_overlap: bool = False,
        seed: int = 0,
        num_jobs: int = 1,
    ) -> CutSet:
        """
        Simulate the desired number of multi-speaker meetings.
        :param cuts: CutSet containing the MonoCut objects to be used for simulation.
        :param num_meetings: the number of meetings to simulate.
        :param num_repeats: the number of times to repeat the provided cuts. This means that
            the number of simulated meetings depends on how many cuts are available.
        :param num_speakers_per_meeting: the number of speakers per meeting. If a list is
            provided, the number of speakers per meeting is sampled from this list.
            [Default: 2]
        :param speaker_count_probs: the probability of each number of speakers per meeting.
            [Default: None]
        :param max_duration_per_speaker: the maximum duration of each speaker in a meeting.
            [Default: 20.0]
        :param max_utterances_per_speaker: the maximum number of utterances per speaker in a
            meeting. [Default: 5]
        :param allow_3fold_overlap: if True, allow 3-fold overlaps between speakers.
            [Default: False]
        :param seed: the random seed to be used for simulation. [Default: 0]
        :param num_jobs: the number of jobs to use for simulation. Use more jobs to speed up
            simulation when you have large number of source utterances. [Default: 1]
        """
        from scipy.stats import bernoulli

        if num_meetings is None and num_repeats is None:
            raise ValueError("Either num_meetings or num_repeats must be provided.")

        if num_meetings is not None:
            num_repeats = None

        if isinstance(num_speakers_per_meeting, int):
            num_speakers_per_meeting = [num_speakers_per_meeting]

        if speaker_count_probs is None:
            speaker_count_probs = [1.0 / len(num_speakers_per_meeting)] * len(
                num_speakers_per_meeting
            )

        # Initialize default distributions if not provided.
        if getattr(self, "same_spk_pause_dist", None) is None:
            self._init_defaults()

        # Create random number generators with the given seed.
        self.bernoulli = bernoulli

        # Create cuts sampler
        sampler = MeetingSampler(
            cuts,
            num_repeats=num_repeats,
            num_meetings=num_meetings,
            max_duration_per_speaker=max_duration_per_speaker,
            max_utterances_per_speaker=max_utterances_per_speaker,
            num_speakers_per_meeting=num_speakers_per_meeting,
            speaker_count_probs=speaker_count_probs,
            seed=seed,
        )
        sampler_iter = iter(sampler)

        work = partial(
            _simulate_worker, simulator=self, allow_3fold_overlap=allow_3fold_overlap
        )

        mixtures = []
        if num_jobs == 1:
            # Don't use multiprocessing if num_jobs == 1.
            for mixture in tqdm(
                map(work, sampler_iter),
                total=num_meetings,
                desc="Simulating meetings",
            ):
                mixtures.append(mixture)
        else:
            for mixture in tqdm(
                parallel_map(
                    work,
                    sampler_iter,
                    num_jobs=num_jobs,
                    queue_size=num_jobs * MAX_TASKS_WAITING,
                ),
                total=num_meetings,
                desc="Simulating meetings",
            ):
                mixtures.append(mixture)

        return CutSet.from_cuts(mixtures)

    def reverberate(self, cuts: CutSet, *rirs: RecordingSet) -> CutSet:
        return reverberate_cuts(cuts, *rirs)


def _simulate_worker(
    utterances, allow_3fold_overlap: bool, simulator: ConversationalMeetingSimulator
):
    return simulator._create_mixture(
        utterances, allow_3fold_overlap=allow_3fold_overlap
    )



from .readers import *
from .writers import *

__all__ = [
    "ArrayTarWriter",
    "AudioTarWriter",
    "JsonlShardWriter",
    "LazySharIterator",
    "SharWriter",
    "TarIterator",
    "TarWriter",
]



from typing import Dict, List, Optional, Tuple, Union

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.speechio import prepare_speechio
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def speechio(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    num_jobs: int = 1,
):
    """SpeechIO data preparation. See https://github.com/SpeechColab/Leaderboard"""
    prepare_speechio(
        corpus_dir=corpus_dir,
        output_dir=output_dir,
    )
